# configutils.py
#
# Written by: Ryan Tanner
# email: ryan.tanner@nasa.gov
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS. If not, see <http://www.gnu.org/licenses/>.

# configutils.py

"""
configutils.py

For configuring defaults of SAS enviroment variables.

"""

# Standard library imports
import os
from configparser import ConfigParser

# Third party imports

# Local application imports

__version__ = 'configutils (configutils-1.0)'

###############
# Configuration

sas_cfg_defaults = {
    "sas_dir": "/does/not/exist",
    "sas_ccfpath": "/does/not/exist",
    "data_dir": "/does/not/exist",
    "verbosity": 4,
    "suppress_warning": 1,
    "on_sci_server": False
}

CURRENT_CONFIG_FILE = None

if (os.path.expanduser("~") == '/home/idies') and \
    os.path.exists('/home/idies/workspace/Storage/'):
    # If on SciServer, get enviroment variables. Replace defaults.
    sas_cfg_defaults['on_sci_server'] = True
    sas_cfg_defaults['usr'] = os.listdir('/home/idies/workspace/Storage/')[0]
    sas_dir     = os.environ.get('SAS_DIR')
    sas_ccfpath = os.environ.get('SAS_CCFPATH')
    sas_cfg_defaults['sas_dir']     = sas_dir
    sas_cfg_defaults['sas_ccfpath'] = sas_ccfpath
    sas_cfg = ConfigParser(sas_cfg_defaults)
    sas_cfg.add_section("sas")
else:
    # If not on SciServer.
    # Check if .config directory exists. If not, make it.
    config_root = os.environ.get(
        "XDG_CONFIG_HOME", os.path.join(os.environ.get('HOME'), ".config")
    )
    CONFIG_DIR = os.path.join(config_root, "sas")
    if not os.path.exists(CONFIG_DIR):
        try:
            os.makedirs(CONFIG_DIR)
        except OSError:
            raise Exception( f'Unable to make SAS config directory: {CONFIG_DIR}')

    # Check if sas.cfg file exists. If not, make it. Populate with defaults.
    CURRENT_CONFIG_FILE = os.path.join(CONFIG_DIR, "sas.cfg")
    if not os.path.exists(CURRENT_CONFIG_FILE):
        cp = ConfigParser(sas_cfg_defaults)
        cp.add_section("sas")
        try:
            with open(CURRENT_CONFIG_FILE, "w") as new_cfg:
                cp.write(new_cfg)
        except IOError:
            raise Exception( f'Unable to write to SAS config file: {CURRENT_CONFIG_FILE}')

    sas_cfg = ConfigParser(sas_cfg_defaults)
    sas_cfg.read([CURRENT_CONFIG_FILE, "sas.cfg"])
    if not sas_cfg.has_section("sas"): sas_cfg.add_section("sas")    

######### Functions #########

def set_sas_config_option(option, value):
    """
    Set default SAS configuration values.

    This sets values as default for future sessions.

    Parameters
    ----------
    option : string
        The option to change.
    value : number or string
        The value to set the option to.
    """
    if os.path.exists(CURRENT_CONFIG_FILE):
        option = option.lower()
        sas_cfg.set("sas", option, value=str(value))
        with open(CURRENT_CONFIG_FILE, "w") as new_cfg:
            sas_cfg.write(new_cfg)
    else:
        print('No SAS configuration file found! Cannot set default value for:')
        print('Option: {0} ; Value: {1}'.format(option,value))

def clear_sas_defaults():
    """
    Clears all SAS defaults set by user. User will have to reset SAS defaults.
            sas_dir
            sas_ccfpath
            data_dir
            verbosity
            suppress_warning
    """
    if os.path.exists(CURRENT_CONFIG_FILE):
        os.remove(CURRENT_CONFIG_FILE)
    else:
        print('SAS configuration file, {0}, not found! Cannot remove!'.format(CURRENT_CONFIG_FILE))




# ESA (C) 2000-2021
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# pysasplot_utils.py

from .version import VERSION, SAS_RELEASE, SAS_AKA

__version__ = f'pysasplot_utils (pysasplot_utils-{VERSION}) [{SAS_RELEASE}-{SAS_AKA}]' 


import pysas.pyutils.pyutils as pyutils
from astropy.io import fits
import matplotlib.pyplot as plt
import numpy as np
import os
from astropy.table import Table
import sys
import re
import warnings
import matplotlib.backends.backend_pdf
import pickle
from pypdf import PdfMerger
from matplotlib.patches import Rectangle
from matplotlib.collections import PatchCollection

def text_plot(fits_file, extra_text = ''):
    """
    Prepares the text for the plot reading Keywords from the given FITS file.

    Args:
        fits_file: the path to the FITS file.
        extra_text: additional text to add to the plot. Empty by default.

    Output:
        text_plot: the text as a string to be added to the plot.
    """

    obs = pyutils.get_key_word(fits_file, 'OBS_ID')
    expo = pyutils.get_key_word(fits_file, 'EXPIDSTR')
    inst = pyutils.get_key_word(fits_file, 'INSTRUME')
    revolut = pyutils.get_key_word(fits_file, 'REVOLUT')
    ra = pyutils.get_key_word(fits_file, 'RA_OBJ')
    dec = pyutils.get_key_word(fits_file, 'DEC_OBJ')
    obj = pyutils.get_key_word(fits_file, 'OBJECT')
    date_obs = pyutils.get_key_word(fits_file, 'DATE-OBS')

    text_plot = 'OBS ID: {}\nInstrument: {}\nExposure: {}\nRevolution: {}\nDate: {}\nRA & DEC: ({} {})\n\n{}'.format(obs, inst, expo, revolut, date_obs, ra, dec, extra_text)

    return text_plot


def ingest_data(fits_file, in_data, label = 'label'):
    """
    Prepares the data from the input in a key-value, ready to plot with the
    other functions.

    Args:
        fits_file: the path or the HDU mobject for the fits file.
        in_data: the data to use. Can by an ndarray object or string-like, as in extension:card.
        label: the label to address the data.

    Output:
        data: the adapted dictionary with the {label:data}
    """

    if isinstance(fits_file, str):
        try:
            f = fits.open(fits_file)
        except (FileNotFoundError, ValueError) as e:
            warnings.warn('Could not locate file {}. Quitting...'.format(fits_file))
            sys.exit(0)
    elif isinstance(fits_file, fits.hdu.hdulist.HDUList):
        f = fits_file
    else:
        warnings.warn('Cannot open the FITS file. Checking the card for array instead of extensions...')

    data = dict()

    if isinstance(in_data, np.ndarray):
        print('Input data: numpy array.')
        data.update({label : in_data})
    elif isinstance(in_data, str):
        in_data = in_data.strip()
        ext, card = in_data.split(':')
        try:
            data.update({ext + ':' + card : f[ext].data[card]})
        except KeyError:
            print('Could not find the combination of extension-key {} in the current FITS file.'.format(ext + ':' + card))
            f.close()
            return None
    elif is_iterable(in_data):
        for comb in range(0, len(in_data)):
            if isinstance(in_data[comb], np.ndarray):
                if is_iterable(label):
                    try:
                        data.update({label[comb] : in_data[comb]})
                    except IndexError:
                        data.update({'{0}_{1}'.format(label[-1], comb) : in_data[comb]})
                else:
                    data.update({'{0}_{1}'.format(label, comb) : in_data[comb]})
            else:
                try:
                    in_data[comb] = in_data[comb].strip()
                    ext, card = in_data[comb].split(':')
                    try:
                        data.update({ext + ':' + card : f[ext].data[card]})
                    except KeyError:
                        print('Could not find the following combination key: {}.'.format(ext + ':' + card))
                        continue
                except KeyError:
                    print('Could not find combination {} in the given FITS file.'.format(ext + ':' + card))
                    continue
    else:
        print('Format unrecognised for x data. Quitting...')
        f.close()
        return None
    
    f.close()

    return data


def is_iterable(py_obj):
    """
    Given a Python object, will return True or False if it's and iterable object aside
    from string.

    Args:
        py_obj: the object that has to be evaluated.

    Output:
        bool: whether or not is iterable.
    """

    from collections.abc import Iterable

    if isinstance(py_obj, Iterable) and not isinstance(py_obj, str):
        return True
    else:
        return False


def get_time_deltas(fits_file, extension = 1, points = 512, x_data = 'TIME', card = 'TIMEDEL'):
    """
    Finds the proper Delta intervals for a FITS file.

    Args:
        fits_file: the fits file: either the path or one already opened.
        extension: the extension for the time sereis to be evaluated. 0 by
    default (primary HDU).
        points: the final number of points wanted. Default: 512.
        x_data: the data to evaluate the x-axis delta. TIME by default.
        card: the card to search in the extension's header. TIMEDEL by default.

    Ouput:
        (DTin, DTout): the initial and final delta time intervals.
        Will return (1, 1) if the file could not be opened or if the extension
    does not contain a proper TIME data column (implying no rebinning).
    """
    
    if not isinstance(fits_file, Table):
        timedel = pyutils.get_key_word(fits_file, card, extension)
    else:
        timedel = 'unknown'

    if isinstance(fits_file, str):
        try:
            with fits.open(fits_file) as f:
                t0 = f[extension].data[x_data][0]
                tf = f[extension].data[x_data][-1]
                fsize = f[extension].data.size
        except FileNotFoundError:
            print('Could not locate file.')
            return (1,1)
    elif isinstance(fits_file, fits.hdu.hdulist.HDUList):
        try:
            t0 = fits_file[extension].data[x_data][0]
            tf = fits_file[extension].data[x_data][-1]
            fsize = fits_file[extension].data.size
        except KeyError:
            print('Could not locate TIME column in data at [{0}].'.format(extension))
            return (1, 1)
    elif isinstance(fits_file, Table):
        t0 = fits_file[x_data][0]
        tf = fits_file[x_data][-1]
        fsize = len(fits_file)

    if timedel == 'unknown':
        DTin = int(np.round(abs(tf - t0) / fsize))
    else:
        DTin = timedel

    DTout = int(np.round((tf - t0) / points))

    return DTin, DTout


def TSrebin(tstable, DTin, DTout, t_xlabel = 'TIME', t_ylabel = 'RATE', t_elabel = 'ERROR', fracexp = 'FRACEXP'):
    """
    Re-binning function.

    Input:
        tstable: time series with TIME, RATE and ERROR columns.
        DTin: the input bin size.
        DTout: the output bin size.
        t_xlabel: the label to search for the x value. TIME by default.
        t_ylabel: the label to search for the y value. RATE by default.
        t_elabel: the label to search for the error values. ERROR by default.
        fracexp: the label to search for the FRACEXP. FRACEXP by default.

    Output:
        ntstable: a Table object with the corrected bin size.
    """

    #
    # Check input Table Columns, in case it's been loaded as a rec array (as in ext.data)
    #

    if isinstance(tstable, fits.fitsrec.FITS_rec):
        tstable = Table(tstable)

    colNames = tstable.keys() # Input ColumnNames
    mandatoryColumns = [t_xlabel, t_ylabel, t_elabel, fracexp] # Mandatory ColumnNames
    diff = set(mandatoryColumns) - set(colNames)

    fracexp_chk = 0
    if diff == {fracexp}:
        print('FRACEXP is missing. fracexp = 1')
        fracexp_chk = 1
    elif len(diff) != 0:
        print('Missing Columns in input Table : ', diff, '\n Exiting.')
        sys.exit(1)

    #
    # Re-Binning of the Table 'tstable'
    #

    tstable.sort(t_xlabel)
    ntime = tstable[t_xlabel]
    nrate = np.nan_to_num(tstable[t_ylabel])
    nerror = np.nan_to_num(tstable[t_elabel])

    if fracexp_chk == 1:
        fracexp = np.linspace(1, 1, len(ntime))
    else:
        fracexp = np.nan_to_num(tstable[fracexp])

    tcum = np.nancumsum(tstable[t_xlabel])
    rcum = np.nancumsum(nrate * fracexp) # Errors propagation are weighted by the FRACEXP from the TS
    ecum = np.nancumsum(nerror**2 * fracexp)
    fcum = np.nancumsum(fracexp)

    nin = len(rcum) # Number of In points

    t_bin_frac = int(np.round(DTout / DTin))

    if t_bin_frac <= 1: # No re-binning
        ntstable = tstable
        new_time_bin = t_bin_frac
        print('    No re-binning needed for this Light Curve. Input / Output time bins ratio : ', t_bin_frac)
    else:
        nout = int(nin/t_bin_frac) # Number of Out points. t_bin_frac == DTout / DTin ('TIMEDEL'). Re-binning factor

        #print('    Input / Output time bins / ratio : ', DTin, 'sec', ' / ',DTout, 'sec', ' / ', t_bin_frac)

        #print('    Input / Output dots number       : ', nin, ' / ',nout)

        i1 = np.arange(nout) * t_bin_frac
        i2 = np.concatenate((i1[1:] - 1, np.array([nin - 1])))

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category = RuntimeWarning) #
            if fracexp_chk == 0: # TS has FRACEXP column => rate and error are weighted by FRACEXP
                nrate  = ((rcum[i2] - rcum[i1]) / (fcum[i2] - fcum[i1])    )
                nerror = (np.sqrt(np.nan_to_num(ecum[i2] - ecum[i1])) / (fcum[i2] - fcum[i1]))
                ntime = (tstable[t_xlabel][i2] + tstable[t_xlabel][i1]) / 2.
                nfrac = (fcum[i2] - fcum[i1]) / (i2 - i1)
                ntstable = Table([ntime, nrate, nerror, nfrac], names = [t_xlabel, t_ylabel, t_elabel, fracexp])
            else: # TS has NOT FRACEXP column => fracexp = 1
                nrate  = ((rcum[i2]-rcum[i1]) / (fcum[i2] - fcum[i1])    )
                nerror = (np.sqrt(np.nan_to_num(ecum[i2] - ecum[i1]))/(fcum[i2]-fcum[i1]))
                ntime = (tstable[t_xlabel][i2] + tstable[t_xlabel][i1])/2.
                ntstable = Table([ntime, nrate, nerror], names = [t_xlabel, t_ylabel, t_elabel])

    return ntstable


def load_figures(list_of_saved_figs):
    """
    Loads a list of pickled figures (passed as a list of paths) and returns a list of figure objects.

    Args:
        list_of_saved_figs: a list containing the paths of the pickled figures.

    Output:
        saved_figs: a list of figures.
    """

    saved_figs = []

    for i in list_of_saved_figs:
        with open(i, 'rb') as f:
            try:
                unpickled = pickle.load(f)
                saved_figs.append(unpickled)
            except AttributeError:
                warnings.warn('Could not open the pickled file: callbacks error.')
                saved_figs.append(None)

    return saved_figs


def saves_in_pdf(fig_list, output_name = 'output', papertype = 'a4'):
    """
    Saves the given figures into a single pdf file.

    Args:
        fig_list: a list or tuple containing the Matplotlib elements to be
    saved into a pdf. Recommended: matplotlib.figure.Figure.
        output_name: the name for the pdf file.
        papertype: the paper type for the output. A4 by default.

    Output:
        the absolute path of the created pdf.
        None if the list is empty or the type of fig_list is not correct.
    """

    from matplotlib.backends.backend_pdf import PdfPages

    try:
        if len(fig_list) == 0:
            warnings.warn('The given list is empty.')
            return None
    except TypeError:
        warnings.warn('The input fig_list is not iterable. Quitting...')
        return None

    missed_fig = 0

    with PdfPages('{}'.format(output_name)) as pdf:
        for i in range(0, len(fig_list)):
            try:
                pdf.savefig(fig_list[i], papertype = papertype)
            except ValueError:
                warnings.warn('Index {0} cannot be saved as it is not recognised as a Figure. Skipping...'.format(i))
                missed_fig = missed_fig + 1
                continue

    if missed_fig == len(fig_list):
        warnings.warn('The pdf is empty due to all figures being non-compatible.')

    return os.path.abspath('{}'.format(output_name))


def simple_1d_hist(y_data, x_label = 'x', y_label = 'y', add_text = '', scale = '', nbins = 20, sharebins = False, \
    fits_file = '', fits_info = True, plot_title = '', outformat = 'pdf'):
    """
    Utility for simple 1 dimention histogram.

    Args:
        y_data: the data for the histogram. Uses the inputs from ingest_data. For example, if fits_file has an input: y_data = ['RATE:RATE', array1, array2...].
        x_label: the label for the x-axis.
        y_label: the label for the y-axis.
        add_text: additional text to add alongside the histogram.
        scale: the desired scale to adapt the y-data.
        nbins: the number of bins for the histogram. 20 by default.
        sharebins: whether or not force all the data to have the same bins distribution.
        fits_file = the fits file to use (if any).
        fits_info: Bool. Whether or not show basic info from the FITS file.
        plot_title: the title of the plot.
        outformat: the format (according to matplotlib. PDF by default.
    """

    plt.style.use('ggplot')
    y_data = ingest_data(fits_file, y_data, y_label)

#    if y_data_errors != '':
#        y_data_errors = ingest_data(fits_file, y_data_errors, y_label)

    fig = plt.figure(figsize = (16, 9), facecolor = 'white')

    if fits_info or add_text != '':
        ax = fig.add_axes([0.3, 0.3, 0.6, 0.4])
    else:
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])

    if sharebins:
        y_data_list = list(y_data.values())
        nbins = np.histogram(np.hstack(y_data_list), bins = nbins)[1]

    for key in y_data.keys():
        try:
            #y_key_error = y_data_errors[key]
            y_error_flag = True
        except (KeyError, TypeError) as e:
            y_error_flag = False
        if not y_error_flag:
            heights, bins, _ = ax.hist(y_data[key], label = str(key), histtype = 'step', fill = True, alpha = 0.2, bins = nbins)
        else:
            heights, bins, _ = ax.hist(y_data[key], label = str(key), histtype = 'step', fill = True, alpha = 0.2, bins = nbins)
            #TODO add errors as separate plotting

    ax.legend(bbox_to_anchor = (1.05, 1), loc = 'upper left', borderaxespad = 0.)

    if fits_file != '' and fits_info:
        add_text = text_plot(fits_file, add_text)

    if add_text != '':
        plt.gcf().text(0.02, 0.5, add_text, fontsize = 14)
    plt.subplots_adjust(top = 0.3)

    if scale != '':
        try:
            plt.yscale(scale)
        except ValueError:
            warnings.warn('The given scale is not supported. Using linear scale instead.')
            plt.yscale('linear')

    if is_iterable(x_label):
        x_label = x_label[0]
    if is_iterable(y_label):
        y_label = y_label[0]

    if plot_title == '':
        plot_title = '{0}_{1}_hist'.format(x_label, y_label)

    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(plot_title)

    # writes the filename filtering out non_alphabetical characters
    file_name = re.sub(r'\W+', '', plot_title)

    # will save the plot. If the program is running in a Notebook environment, will show the plot.
    if pyutils.is_notebook():
        plt.show()
        if outformat.upper() == 'PDF':
            saves_in_pdf([fig], file_name)
        elif outformat.upper() == 'PNG':
            plt.savefig(file_name + '.png')
        else:
            warnings.warn('Unrecognised format, will not be saved.')
    else:
        if outformat.upper() == 'PDF':
            saves_in_pdf([fig], file_name)
        else:
            try:
                plt.savefig(file_name + '.{}'.format(outformat))
            except ValueError:
                warnings.warn('Unrecognised format, will not be saved.')

    return fig


def simple_2d_plot(x_data, y_data, x_data_errors = '', y_data_errors = '', plot_title = '', x_label = 'x', y_label = 'y', fits_file = '', fits_info = True, \
    add_text = '', scale = '', points = 512, outformat = 'png'):
    """
    Utility for simple 2 dimentional plots. The data can be:
        from FITS files, written as in EXT:CARD.
        iterable types, including numpy arrays.
        Iterable types with a mix of the above.

    If using a FITS file, always use the phrasing EXT:CARD to refer to an
    extension and card data. Introduce first this information and leave the
    other type of arrays for last.

    Args:
        x_data: the x coordiantes of the data.
        y_data: the y coordinates of the data.
        x_data_errors: errors for the x axis.
        y_data_errors: errors for the y axis.
        plot_title: title for the plot. Will be used also when writting the file.
        x_label: label for the x axis. x by default.
        y_label: label for the y axis. y by default.
        fits_file: only needed if requesting an extension:card in one of the data inputs.
        add_text: additional text to add. None by default.
        scale: the scale to apply.
        points: the number of points. Required in case of rebinning.
        outformat: the format for the output.
    """

    plt.style.use('ggplot')


    if isinstance(fits_file, Table):
        x_data, y_data, x_data_errors, y_data_errors = ingest_table(fits_file, points, x_data, y_data, x_error_label = x_data_errors, y_errors_label = y_data_errors)
        fits_info = False # it's impossible to extract those values from a table object
    else:
        x_data = ingest_data(fits_file, x_data, x_label)
        y_data = ingest_data(fits_file, y_data, y_label)

        if y_data_errors != '':
            y_data_errors = ingest_data(fits_file, y_data_errors, y_label)
        if x_data_errors != '':
            x_data_errors = ingest_data(fits_file, x_data_errors, x_label)

    fig = plt.figure(figsize = (16, 9), facecolor = 'white')
    if fits_info or add_text != '':
        ax = fig.add_axes([0.3, 0.3, 0.6, 0.4])
    else:
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])

    if len(list(x_data.keys())) == 1:
        x_data_key = list(x_data.keys())[0]
        x_data = list(x_data.values())[0]

        for key in y_data.keys():
            try:
                y_key_error = y_data_errors[key]
                y_error_flag = True
            except (KeyError, TypeError) as e:
                y_error_flag = False
            try:
                x_key_error = x_data_errors[key]
                x_error_flag = True
            except (KeyError, TypeError) as e:
                x_error_flag = False
            if not y_error_flag and not x_error_flag:
                try:
                    ax.plot(x_data, y_data[key], marker='.', label = str(key), linewidth = 0.8)
                    ax.legend(bbox_to_anchor = (1.05, 1), loc = 'upper left', borderaxespad=0.)
                except ValueError:
                    print('Mismatching sizes for x and y: {} and {}. Omitting this curve.'.format(len(x_data), y_data[key].size))
                    continue
            elif x_error_flag and not y_error_flag:
                try:
                    ax.errorbar(x_data, y_data[key], xerr = x_key_error, marker='.', label = str(key), linewidth = 0.8)
                except ValueError:
                    print('Mismatching between the given arrays. Key: {}.'.format(key))
                    continue
            elif y_error_flag and not x_error_flag:
                try:
                    ax.errorbar(x_data, y_data[key], yerr = y_key_error, marker='.', label = str(key), linewidth = 0.8)
                except ValueError:
                    print('Mismatching between the given arrays. Key: {}.'.format(key))
                    continue
            elif y_error_flag and x_error_flag:
                try:
                    ax.errorbar(x_data, y_data[key], xerr = x_key_error, yerr = y_key_error, marker='.', label = str(key), linewidth = 0.8)
                except ValueError:
                    print('Mismatching between the given arrays. Key: {}.'.format(key))
                    continue

    if fits_info and (isinstance(fits_file, str) or isinstance(fits_file, fits.hdu.hdulist.HDUList)):
        add_text = text_plot(fits_file, add_text)

    plt.gcf().text(0.02, 0.5, add_text, fontsize=14)
    #plt.subplots_adjust(top=0.3)

    if scale != '':
        try:
            plt.yscale(scale)
        except ValueError:
            warnings.warn('The given scale is not supported. Using linear scale instead.')
            plt.yscale('linear')

    if is_iterable(x_label):
        x_label = x_label[0]
    if is_iterable(y_label):
        y_label = y_label[0]

    if plot_title == '':
        plot_title = '{0}_{1}_plot'.format(x_label, y_label)

    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(plot_title)

    file_name = re.sub(r'\W+', '', plot_title)
    # will save the plot. If the program is running in a Notebook environment, will show the plot.
    if pyutils.is_notebook():
        plt.show()
        if outformat.upper() == 'pdf':
            saves_in_pdf([fig], file_name)
        else:
            plt.savefig(file_name + '.{}'.format(outformat))
    else:
        if outformat.upper() == 'pdf':
            saves_in_pdf([fig], file_name)
        else:
            try:
                plt.savefig(file_name + '.{}'.format(outformat))
            except ValueError:
                warnings.warn('Unrecognised format. Will not save plot.')

    return fig


def ingest_table(table, points, x_label, y_labels, x_error_label = None, y_errors_label = None):
    """
    Ingest the data from a given table and returns the data using Python
    dictionaries with labels. Only labels from the table are supported.

    Args:
        table: the table object.
        points: the number of points wanted. A rebinning will be applied to the table.
        x_label: the label for the x axis.
        y_labels: the label (or array of labels) for the y data.
        x_error_label: the label for the error for the x-coordinates.
        y_error_labels: the label (or array of labels) for the error of the
    y-coordinates.

    Output:
        (x_data, y_data, x_error, y_data_errors): set of dictionaries containing
    the information from the table.
    """

    DTin, DTout = get_time_deltas(table, 1, points)
    t_bin_frac = int(np.round(DTout / DTin))
    if t_bin_frac <= 1:
        print('No rebinning.')
    else:
        table = TSrebin(table, DTin, DTout)

    x_data = {x_label : table[x_label]}
    y_data = dict()

    if is_iterable(y_labels):
        for label in y_labels:
            y_data.update({label : table[label]})
    elif isinstance(y_labels, str):
        y_data.update({y_labels : table[y_labels]})

    nrows = len(table)
    x_error = dict()

    try:
        if x_error_label:
            x_error = x_error.update({x_error_label : table[x_error_label]})
    except KeyError:
        print('Could not find errors for the x axis in {0}.'.format(x_error_label))

    y_error = dict()

    if y_errors_label:
        if is_iterable(y_errors_label):
            for label in y_errors_label:
                try:
                    y_error.update({label : table[label]})
                except KeyError:
                    print('Could not find {0} in the given table.'.format(label))
                    continue
        elif isinstance(y_errors_label, str):
            try:
                y_error.update({y_labels : table[y_labels]})
            except KeyError:
                print('Could not find {0} in the given table.'.format(y_labels))

    return(x_data, y_data, x_error, y_error)


def merge_pdf(pdf_files, output_file):
    """
    Merges several PDF files passed as the arguments. 
    
    Args:
        pdf_files: list or tuple containing the paths to the pdfs.
        output_file: the name of output file.

    Output:
        1 depending on whether or not the process has been completed; 0 if not.
    """

    merger = PdfMerger()
    total_files = len(pdf_files)
    e_count = 0

    for pdf in pdf_files:
        try:
            merger.append(pdf)
        except:
            print('Could not append {}.'.format(pdf))
            e_count = e_count + 1

    with open(output_file, 'wb') as out:
        merger.write(out)

    if e_count == total_files:
        print('ERROR: No pdf was merged')
        return 0

    #merger.write(output_file)
    
    merger.close()

    return 1


def plot_image(path_to_image, ext = 0, output = None):
    """
    Plots and saves an image corresponding to a FITS image.

    Args:
        path_to_image: the path to the FITS file.
        ext: the extension of the image in the FITS file. 0 by default.
        output: None by default. Whethere or not save the image.

    Output:
        0 when finished.
    """

    with fits.open(path_to_image) as img:
        img_data = img[ext].data

    plt.imshow(img_data)

    if output is None:
        if os.getenv('DISPLAU'):
            plt.show()
        else:
            warnings.warn('Nothing was found in DISPLAY -- nothing was produced')
        return 0
    else:
        plt.savefig(output)

    return 0


def check_format_compatibility(out_format):
    """
    Checks if the input format is available as a matplotlib backend format.

    Args:
        out_format: the format to be checkec.

    Output:
        bool: whether or not the format can be used.
    """

    format_list = ('PNG', 'PDF', 'PGF', 'EPS', 'SVG2', 'RGBA', 'RAW', 'PS', 'SVG')

    if out_format.upper() in format_list:
        return True
    else:
        return False


def plot_region_box(ax, x1, y1, width, height, angle = 0, fill = False, colour = 'green', text = None, transform = None):
    '''
    Returns an axis object with with a box region printed on it.

    Args:
        ax: the original axis to use for the plot.
        x1: the starting x coordinate for the box.
        y1: the starting y coordinate for the box.
        width: the width of the box.
        height: the heigth of the box.
        angle: the angle of the box (if any). 0 by default.
        fill: boolean. Whether or not fill the box.
        colour: the colour of the box and the fill (if fill = True).
        text: None by default. Text to add at the (x1, y1) coordinates.
        transform: None by default. The transformation for the axis. Must be comatible with matplotlib standards.

    Output:
        f_ax: the axis after plotting the box.
    '''
     
    rec = Rectangle((x1, y1), width, height, angle, fill = fill, transform = transform, color = colour)
    
    pc = PatchCollection([rec], match_original = True)
    

    if text is None:
        pass
    else:
        ax.annotate(text, (x1, y1), color = colour, fontsize = 10)
        
    ax.add_collection(pc)
    
    return None


def reg_to_list(reg):
    ''' 
    Gets the information from a region based on the input string (based on ds9 formatting).

    So far, only tested for BOX regions.

    Args:
        reg: the region string in ds9 standards.

    Output:
        TODO: must be useful for all shapes. 
    ''' 
    
    shape_info, shape_details = reg.split('#') 
    shape_info = shape_info.strip().rstrip() 
    
    if 'BOX' in shape_info.upper(): 
        polygon = 'BOX' 
        x1, y1, width, heigth, angle = re.findall(r"[-+]?(?:\d*\.\d+|\d+)", shape_info)
        shape_info = (polygon, x1, y1, width, heigth, angle)
        
    # text details 
    if shape_details.find('COLOR') == -1: 
        colour = 'green' 
    else: 
        c_letters = shape_details[shape_details.find('COLOR') + 6 : shape_details.find('COLOR') + 9]
        if c_letters.upper() in 'RED': 
            colour = 'red' 
        elif c_letters.upper() in 'GREEN': 
            colour = 'green' 
        elif c_letters.upper() in 'BLUE': 
            ccolour = 'blue' 
        elif c_letters.upper() in 'CYAN': 
            colour = 'cyan' 
        elif c_letters.upper() in 'WHITE':  
            ccolour = 'white' 
        elif c_letters.upper() in 'MAGENTA': 
            colour = 'magenta' 
        elif c_letters.upper() in 'YELLOW': 
            ccolour = 'yellow' 
        else: 
            warnings.warn('Colour not recognised. Using green by default') 
            colour = 'green' 

    if not 'TEXT' in shape_details.upper(): 
        text = None
    else: 
        text = shape_details[shape_details.find('TEXT') + 6: shape_details.find('}')]
            
            
    return (polygon, float(x1), float(y1), float(width), float(heigth), float(angle), colour, text)



# ESA (C) 2000-2021
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#
""""runtask.py

It will run a Python task which has been coded into a module
named taskname.py.

It is assumed that taskname.py is part of package taskname which
will be placed below pysas.

iparsdic is a dictionary which includes all task parameters (as read
from par file) and their defaults except those modified in the
command line in the form param=value pairs.

iparsdic is passed to the task module, which will be taken in its
run function defined there.
"""

# Standard library imports
from importlib import import_module
import pkgutil
import os, sys
import subprocess

# Third party imports

# Local application imports



class RunTask:
    """Class RunTask
    
    Parameters are entered in dictionary iparsdic.

    Method run executes the task depending on whether
    is a SAS native task or a Python task.
    
    """

    def __init__(self, taskname, iparsdic,logFile):
        self.taskname = taskname
        self.iparsdic = iparsdic
        self.logFile = logFile

    def run(self):
        """"Method run

            If taskname is a Python module, therefore it is in the
            list pysaspkgs, then import it and pass to its run
            function the dictionary of parameters, iparsdic

            If taskname is not a Python SAS task, there will not be
            a run function, so we will invoke subprocess
            """

        sas_path = os.environ.get('SAS_PATH')

        if not sas_path:
            raise Exception('SAS_PATH is undefined! SAS not initialised?')

        saspath = sas_path.split(':')

        sasdev = saspath[0]

        pysasdevdir = os.path.join(sasdev, 'lib', 'python', 'pysas')

        pysaspkgs = []

        for p in pkgutil.walk_packages([pysasdevdir]):
            if p.ispkg:
                pysaspkgs.append(p.name)


        if self.taskname in pysaspkgs:

            m = import_module('pysas.' + self.taskname + '.' + self.taskname)

            m.run(self.iparsdic)

        else:
            cmd = ''
            cmd += self.taskname + ' '
            for parval in self.iparsdic.items():
                k, v = parval
                #Remove single quotes a double quotes from the python input parameters
                #SOC-SPR-7684
                if v.startswith("\"") or v.endswith("\""):
                    v = v.replace('"','')
                if v.startswith("'") or v.endswith("'"):
                    v = v.replace('\'','')
                
                if ' ' or '|' in v:
                    cmd += k + '=' + "'"
                    vc = v.split(' ')
                    for i in range(len(vc)):
                        if i == len(vc) - 1:
                            cmd += vc[i]
                        else:
                            cmd += vc[i] + ' '
                    cmd += "' "
                else:
                    cmd += k + '=' + v + ' '
                
            print(f'Executing: \n{cmd}')
            self.p= None
            self.stdoutFile = None

            if self.logFile == 'DEFAULT':
                with subprocess.Popen(cmd,
                                      bufsize=1,
                                      shell=True,
                                      text=True,
                                      stdout=subprocess.PIPE,
                                      stderr=subprocess.STDOUT,
                                      universal_newlines=True) as p:
                    for line in p.stdout:
                        print(line, end='')
            else:
                self.stdoutFile=open(self.logFile,'w')
                p = subprocess.Popen(cmd, 
                                     bufsize=1,
                                     shell=True,
                                     text=True,
                                     stdout=self.stdoutFile,
                                     stderr=subprocess.STDOUT,
                                     universal_newlines=True)                              
                p.wait()
                # for line in p.stdout:
                #     self.stdoutFile.write(line)                   
            
            if p.returncode != 0:
                if self.logFile != 'DEFAULT':
                    self.stdoutFile.close()
                raise subprocess.CalledProcessError(p.returncode, p.args)

            if self.logFile != 'DEFAULT':
                self.stdoutFile.close()



#! /usr/bin/env python
#
# ESA (C) 2000-{creationyear}
#
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#
# write_version.py
#
# Writes version.py from $SAS_DIR/SasVersion.h.
# Values are written as constants to be used later on by importing them
# from version.py.
# The version of the package is also written as VERSION

# Standard library imports
import os, re
from datetime import date


# current date and year in string format
today = date.today()
creationyear = today.year
creationdate = today.strftime('%d-%b-%Y')
creationdate_chlog = today.strftime('%Y-%m-%d')


# readSasVersion reads the SAS release information from SasVersion.h

def readSasVersion(sas_dir):

    sasversion_h = {}
    sasversion_h_keys = [
            'sas_compilation_date',
            'sas_compilation_host',
            'sas_compilation_user',
            'sas_platform',
            'sas_release',
            'sas_aka'
            ]
    with open(os.path.join(sas_dir, 'SasVersion.h')) as inf:
        lines = inf.readlines()

    for key in sasversion_h_keys:
        for line in lines:
            if key in line:
                # [^"] = match any characterxs excluding "
                # ([^"]*) = match any text excluding "
                # '"([^"]*)"' = any text between " and "
                k = re.findall(r'"([^"]*)"', line)
                sasversion_h[key] = k[0]
    return sasversion_h


# main
if __name__ == '__main__':
    sas_dir = os.environ.get('SAS_DIR')
    if sas_dir == None:
        print('Error: sasver : SAS_DIR is undefined')
        sys.exit()
    
    pkgdir = '../../'

    with open(os.path.join(pkgdir, 'VERSION')) as f:
        v = f.readline().strip()

# reads the SasVersion.h

    svh = readSasVersion(sas_dir)

# writes version.py taking version.py.in as template

    with open('version.py.in') as inf, open('version.py', 'w') as outf:
        template = inf.read()
        outf.write(template.format(
            creationdate=creationdate,
            creationyear=creationyear,
            release=svh['sas_release'], 
            aka=svh['sas_aka'], 
            date=svh['sas_compilation_date'], 
            host=svh['sas_compilation_host'],
            user=svh['sas_compilation_user'],
            platform=svh['sas_platform'],
            version=v)
            )
    inf.close()
    outf.close()



# ESA (C) 2000-2021
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

"""wrapper.py

This is a wrapper for SAS tasks which are not based on Python.
This code can be used from Jupyter Notebooks and ipython interactive
sessions
"""

# Standard library imports
import subprocess

# Third party imports

# Local application imports
from pysas.sastask import MyTask


class Wrapper:
    """Class Wrapper

    Method run instantiates MyTask to read the parameter file,
    process the input arguments and run the task.
    """
    def __init__(self, taskname, inargs,logFile='DEFAULT'):
        self.taskname = taskname
        self.inargs = inargs
        self.logFile = logFile

    def run(self):
        t = MyTask(self.taskname, self.inargs,self.logFile)
        t.readparfile()
        t.processargs()
        t.runtask()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 13 12:20:26 2023

@author: ryan.tanner@nasa.gov

Script to replicate the XMM-Newton ABC Guide, chapter 6.
https://heasarc.gsfc.nasa.gov/docs/xmm/abc/node8.html
Part 1
"""

# pySAS imports
import gofpysas as pysas
from gofpysas.wrapper import Wrapper as w

# Useful imports
import os

# Imports for plotting
import matplotlib.pyplot as plt
from astropy.visualization import astropy_mpl_style
from astropy.io import fits
from astropy.wcs import WCS
from astropy.table import Table
plt.style.use(astropy_mpl_style)

# Make FITS image file and plot
def make_fits_image(event_list_file, image_file='image.fits'):
    
    inargs = {}
    inargs['table']        = '{0}'.format(event_list_file)
    inargs['withimageset'] = 'yes'
    inargs['imageset']     = '{0}'.format(image_file)
    inargs['xcolumn']      = 'X'
    inargs['ycolumn']      = 'Y'
    inargs['imagebinning'] = 'imageSize'
    inargs['ximagesize']   = '600'
    inargs['yimagesize']   = '600'

    w('evselect', inargs).run()

    hdu = fits.open(image_file)[0]
    wcs = WCS(hdu.header)

    ax = plt.subplot(projection=wcs)
    plt.imshow(hdu.data, origin='lower', norm='log', vmin=1.0, vmax=1e2)
    ax.set_facecolor("black")
    plt.grid(color='blue', ls='solid')
    plt.xlabel('RA')
    plt.ylabel('Dec')
    plt.colorbar()
    plt.show()
    
# Make light curve file and plot
def plot_light_curve(event_list_file, light_curve_file='ltcrv.fits'):
                     
    inargs = ['table={0}'.format(event_list_file), 
              'withrateset=yes', 
              'rateset={0}'.format(light_curve_file), 
              'maketimecolumn=yes', 
              'timecolumn=TIME', 
              'timebinsize=100', 
              'makeratecolumn=yes']

    w('evselect', inargs).run()

    ts = Table.read(light_curve_file,hdu=1)
    plt.plot(ts['TIME'],ts['RATE'])
    plt.xlabel('Time (s)')
    plt.ylabel('Count Rate (ct/s)')
    plt.show()

# 6.1 : Rerun basic processing
obsid = '0123700101'
odf = pysas.odfcontrol.ODFobject(obsid)
odf.basic_setup(overwrite=False,repo='heasarc',
                rerun=False, epproc_args=['withoutoftime=yes'])

# Event list file names and paths are in the odf object
file_names = list(odf.files.keys())
print(file_names)
for name in file_names: print(odf.files[name])

# The data, odf, and work directory paths are in the odf object
print("Data directory: {0}".format(odf.data_dir))
print("ODF  directory: {0}".format(odf.odf_dir))
print("Work directory: {0}".format(odf.work_dir))


# 6.2 : Plot unfiltered image

os.chdir(odf.work_dir)
mos1 = odf.files['m1evt_list'][0]
make_fits_image(mos1)


# 6.3 : Apply Standard Filter
# A lot of that is just background noise. We need to filter it out.
filtered_event_list = 'mos1_filt.fits'

inargs = ['table={0}'.format(mos1), 
          'withfilteredset=yes', 
          "expression='(PATTERN <= 12)&&(PI in [200:4000])&&#XMMEA_EM'", 
          'filteredset={0}'.format(filtered_event_list), 
          'filtertype=expression', 
          'keepfilteroutput=yes', 
          'updateexposure=yes', 
          'filterexposure=yes']

w('evselect', inargs).run()

# Create image file from filtered event list "outfilter"
make_fits_image(filtered_event_list)


# 6.4 : Create Light Curve
# Helps us see if there is particle flaring so we can filter it out.

light_curve_file='mos1_ltcrv.fits'
plot_light_curve(filtered_event_list, light_curve_file=light_curve_file)


# 6.5 : Applying Time Filters
# 6.5.1 : Filter on RATE with tabgtigen

gti_rate_file = 'gti_rate.fits'
mos1_filt_rate = 'mos1_filt_rate.fits'

inargs = ['table={0}'.format(light_curve_file), 
          'gtiset={0}'.format(gti_rate_file),
          'timecolumn=TIME', 
          "expression='(RATE <= 6)'"]

w('tabgtigen', inargs).run()

inargs = ['table={0}'.format(filtered_event_list),
          'withfilteredset=yes', 
          "expression='GTI({0},TIME)'".format(gti_rate_file), 
          'filteredset={0}'.format(mos1_filt_rate),
          'filtertype=expression', 
          'keepfilteroutput=yes',
          'updateexposure=yes', 
          'filterexposure=yes']

w('evselect', inargs).run()

# Create new image file from filtered event list.
# Point sources are left.

make_fits_image(mos1_filt_rate, image_file='final_image1.fits')

# Create new light curve and plot.

plot_light_curve(mos1_filt_rate)


# 6.5.2 : Filter on TIME with tabgtigen

gti_time_file = 'gti_rate.fits'
mos1_filt_time = 'mos1_filt_time.fits'

inargs = ['table={0}'.format(light_curve_file), 
          'gtiset={0}'.format(gti_time_file),
          'timecolumn=TIME', 
          "expression='(TIME <= 73227600)&&!(TIME IN [7.32118e7:7.3212e7])&&!(TIME IN [7.32204e7:7.32206e7])'"]

w('tabgtigen', inargs).run()

inargs = ['table={0}'.format(filtered_event_list),
          'withfilteredset=yes', 
          "expression='GTI({0},TIME)'".format(gti_time_file), 
          'filteredset={0}'.format(mos1_filt_time),
          'filtertype=expression', 
          'keepfilteroutput=yes',
          'updateexposure=yes', 
          'filterexposure=yes']

w('evselect', inargs).run()

make_fits_image(mos1_filt_time, image_file='final_image2.fits')
plot_light_curve(mos1_filt_time)


# 6.5.3 : Filter on TIME with gtibuild

gti_lines = ['0        73227600 + # Good time from the start of the observation',
             '73211800 73212000 - # But without a small flare here.',
             '73211800 73212000 - # And here.']

with open('gti.txt', 'w') as f:
    f.writelines(gti_lines)
    
gti_txt_file = 'gti.txt'
new_gti_file = 'new_gti.fits'
mos1_new_gti = 'mos1_new_gti.fits'

inargs = ['file={0}'.format(gti_txt_file),
          'table={0}'.format(new_gti_file)]

w('gtibuild', inargs).run()

inargs = ['table={0}'.format(filtered_event_list),
          'withfilteredset=yes', 
          "expression='GTI({0},TIME)'".format(new_gti_file), 
          'filteredset={0}'.format(mos1_new_gti),
          'filtertype=expression', 
          'keepfilteroutput=yes',
          'updateexposure=yes', 
          'filterexposure=yes']

w('evselect', inargs).run()

make_fits_image(mos1_new_gti, image_file='final_image3.fits')
plot_light_curve(mos1_new_gti)


# 6.5.4 : Filter on TIME by Explicit Reference

expression = "expression='(PATTERN <= 12)&&(PI in [200:12000])&&#XMMEA_EM&&(TIME <= 73227600) &&!(TIME IN [7.32118e7:7.3212e7])&&!(TIME IN [7.32204e7:7.32206e7])'"

full_filt_event_list = 'mos1_filt.fits'

inargs = ['table={0}'.format(mos1), 
          'withfilteredset=yes', 
          expression, 
          'filteredset={0}'.format(full_filt_event_list), 
          'filtertype=expression', 
          'keepfilteroutput=yes', 
          'updateexposure=yes', 
          'filterexposure=yes']

w('evselect', inargs).run()

make_fits_image(full_filt_event_list, image_file='final_image4.fits')
plot_light_curve(full_filt_event_list)





# ESA (C) 2000-2021
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#
"""sastask.py

Defines a template SASTask Python class
based on the ABC (Abstract Base Classes) module.

All SAS Pyhton based tasks will derive from SASTask.
However, they will not instantiate it but instead a
subclass of it, MyTask. There could be other similar
to MyTask, but for the time being we simply define this one.

The methods defined in SASTask (decorated with
the @abstractmethod decorator), must be redefined in MyTask.
Otherwise, the instantiation of MyTask will fail.
By particularizing the methods defined in SASTask
we ensure all SAS Python based tasks have unified methods.

As any other SAS task, Python based tasks are used to
perform specific data processing jobs.

The task invocation command may include several parameters
and generic options.

Parameters allow to set or to choose among specific task processing
options. SAS parameter syntax has the form param=value.
Mandatory parameters must be present in the command line.
Optional parameters might have a default value.
All task parameters with their properties
must be defined in a parameter file, which is unique for each task.
Such file provides for each paramater, its name, type (int, real,
boolean, string, etc), options (one or several choices), whether it is
mandatory or not, its default value if none is given, etc.
The parameter file is written in XML format.

Generic options, common to all SAS tasks, allow to handle
several features of the execution environment.
They can be expressed either with a single or with a double "-",
e.g. -v or --version (to show the task version and exit), -V or
--verbosity (to set the level of output verbosity between 0 and 10), etc.
Some of these options modify shell environment variables, e.g.
SAS_VERBOSITY is modified by the value given to --verbosity, etc.

Let be mytask a new Python based task. It will be an instance of MyTask
 
mytask = MyTask('mytask', args)
  
where args is a list that includes all arguments we pass to mytask
in the command line. As we will see this is not the only way to
pass arguments to the task.

mytask mandatory files and directories will be created by pkgmaker,
a Python script responsible of creating any type of SAS task, including
also those based on C++, Fortran 90 and Perl.

The specific Python code for mytask, and possibly othe auxiliary Python
code which might be required by mytask to operate, will be placed all
within the directory structure of mytask.

A typical invocation of mytask could be of the form

mytask param0=value0 [param1=value1] ... [Options]

where param0, param1, etc, are any task parameters defined in
a file named mytask.par, and [Options] are any of the generic options
common to all SAS task. If a parameter, e.g. param0, is defined in
mytask.par as 'mandatory', it must be present in the command line.

Of course, the presence of some of the Options, will trigger specific
immediate actions: e.g. the option -v or --version will immediately
show the version of mytask and exit.
 
Arguments of the form param=value can alternate with Options. 
The only requirement is that Options with a value, e.g. -V 4, must be
adjacent.

Methods defined by SASTask (identified so far):

1. readparfile: Read and loads the task parameter file.

2. processargs : Parses arguments and acts accordingly. 

3. runtask: Executes the task with the proper arguments.

MyTask will actually implement these three methods.

The readparfile method used class paramXmlInfoReader from
param.py module.

The processargs methdos uses the class ParseArgs based on argparse.
 
The runtask method uses class RunTask.
"""

# Standard library imports
from abc import ABC, abstractmethod
import os, numbers

# Third party imports

# Local application imports
from pysas.param import paramXmlInfoReader
from pysas.parser import ParseArgs
from pysas.runtask import RunTask


# Class SASTask

class SASTask(ABC):
    """This is the abstract base class for all SAS Python tasks"""

    def __init__(self, taskname, inargs):
        self.name = taskname
        self.inargs = inargs

    def __repr__(self):
        return f'{self.__class__.__name__}({self.taskname} - {self.inargs})'

    @abstractmethod
    def readparfile(self):
        pass

    @abstractmethod
    def processargs(self):
        pass

    @abstractmethod
    def runtask(self):
        pass



class MyTask(SASTask):
    """
    Class MyTask is a children of SASTask which 
    implements all its abstract methods.

    In the class initialization, the task name and
    the input args to run it are passed from the 
    instatiation of MyTask. Task parameters as
    identified by the '=' sign are separated of 
    task options, to reorder them to avoid conflicts 
    at the time of parsing them.

    The instance method 'readparfile' gets a full
    picture of the task parameter file, receiving
    a bunch of information abouth subparameters and
    their relationship with its predecessors.

    The instance method 'processargs' performs the
    processing of any immediate options and filter out
    the legitimate and mandatory parameters so as
    they can be used in the ' runtask' instance method.
    """

    def __init__(self, taskname, inargs,logFile='DEFAULT'):
        super().__init__(taskname, inargs)
        self.taskname = taskname
        self.inargs = inargs
        self.logFile = logFile

        # Check if inargs is a 'dict'. If it is then convert to list format.
        if isinstance(self.inargs, dict):
            # Get dict of default inputs for the task.
            t = paramXmlInfoReader(self.taskname)
            t.xmlParser()
            defdict = t.defaultValues()
            defkeys = defdict.keys()
            inkeys = self.inargs.keys()
            outparams = []
            
            # Loop over all keys in the inargs dict and add values in 
            # the outdict, but only if different from default values. 
            # Build outparams.
            for key in list(inkeys):
                if key == 'options':
                    # Only if 'options' is not empty.
                    if self.inargs[key] != '':
                        outparams.append(self.inargs[key])
                else:
                    # Safety check to see if a number was passed in.
                    if isinstance(self.inargs[key], numbers.Number):
                        self.inargs[key] = str(self.inargs[key])
                    outparams.append(key+'='+self.inargs[key])
            self.inargs = outparams

        # Reorder self.inargs to group together all options and 
        # all args of type param=value, in that order.
        sasparams = []
        options = []
        for a in self.inargs:
            if '=' in a.split('=',1):
                sasparams.append(a)
            else:
                options.append(a)
        self.inargs = [*options, *sasparams]

    def __repr__(self):
        return f'{self.__class__.__name__}({self.taskname} - {self.inargs})'

    def readparfile(self):
        t = paramXmlInfoReader(self.taskname)
        t.xmlParser()
        self.allparams = t.allparams
        self.mandparams = t.mandpar
        self.mainparams = t.mainparams
        self.parmap = t.parmap
        self.mandpar_dict = t.mandpar_dict
        self.rev_mandpar_dict = t.rev_mandpar_dict
        self.rev_mandpar_string_dict = t.rev_mandpar_string_dict


    def processargs(self):
        p = ParseArgs(self.taskname, self.inargs)
        p.taskparser()
        # tparams is a list with the left hands of param=value, if any
        self.tparams = p.tparams

        # Execute options which require immediate action
        # Options which are for environment (which are cumulative) 
        # do not change default False value of Exit.
        self.Exit = p.procopt()
        if self.Exit:
            return self.Exit

        # 1st check: Whether or not parameters in inargs are defined in the parameter file
        for p in self.tparams:
            if p.strip() not in self.allparams.keys():
                raise Exception(f'Parameter {p} is not defined in the parameter file')

 
        # 2nd check: Whether we have mandatory parameters and subparameters
        for k, v in self.rev_mandpar_dict.items():
            if k.strip() in v and k.strip() not in self.tparams:
                raise Exception(f'Missing, at least, mandatory parameter {k}')
            if k not in v:
                for p in self.tparams:
                    if p in v:
                        implicit = 'yes'
                        for c in v:
                            if c not in self.tparams:
                                implicit = 'no'
                                raise Exception(f'Besides {p}, mandatory subparameter {c} must also be present')
                        if implicit == 'yes':
                            defval = self.allparams[k]['default']
                            if self.allparams[k]['type'] == 'string':
                                self.rev_mandpar_string_dict[k].remove(defval)
                                alt = self.rev_mandpar_string_dict[k].pop()
                            elif self.allparams[k]['type'] == 'bool' and defval == 'no':
                                alt = 'yes'
                            elif self.allparams[k]['type'] == 'bool' and defval == 'yes':
                                alt = 'no'
                            newinarg = k + '=' + alt
                            self.inargs.append(newinarg)
                            if k in self.tparams:
                                print(f'Warning: No need to include {k}. Assumed {newinarg}')
                            break
                    elif p in v and k not in self.tparams:
                        raise Exception(f'Mandatory sub-parameter {p} requires {k} be present')

        # If any subparameters of a parent parameter are set in the command
	    # line or the arguments list, the parent can not keep its default
	    # value. This is typical for parent parameters of type 'boolean' whose
	    # default value is 'no' (or 'yes'). In such a case the parent value must 
        # be changed to 'yes' (or 'no'). This is known as 'implicit' behaviour. 
	    # The previous code takes this behaviour in consideration but
	    # only for mandatory parameters. Now it will be applied to all
	    # parameters having subparameters. 
	    # The final effect of this will be visible below when merging 
	    # the parsdic (the dictionary with the default values for 
	    # all the parameters) and the argsdic to pass all task  
	    # parameters for execution.

	    # Now compute a dictionary for all implicit parameters. 
	    # keys are the parent parameters having any subparameters and
	    # values are lists with these subparameters as elements.
	    # Neither the keys nor the elements of the lists can be mandatory.

        implicitparams = {}
        for d in self.parmap:
            for k, v in d.items():
                if v != [] and k not in self.mandparams:
                    for sp in v:
                        if sp not in self.mandparams:
                            implicitparams[k] = v

        # At this point in inargs there should be only optional modifiers
        # and true task parameters

        # Given that we know all parameters for the task (self.allparams) with
        # their default values, and the parameters and their values entered 
        # in the command line, let us produce now a single object with all 
        # parameters which we can pass to the module to run.

        # argsdic is a dictionary with the pairs param, value 
        # as entered from the command line from 'param=value'

        argsdic = {}
        for a in self.inargs:
            if '=' in a:
                k, v = a.split('=', 1)
                argsdic[k] = v

        # Load defaults with all parameters default values.
        # Use dictionary method 'setdefault' to set the value for a given key;
        # here the key is 'default'.
        # If the value for that key is not defined, fill it with ''.

        defaults = {}
        for a in self.allparams.values():
            defaults[a['id']] = a.setdefault('default', '')

        # 3rd Check: Whether any subparameter is set in argsdic or not.
        # We assume the parent is boolean ('yes'/'no').

        for a in argsdic.keys():
            for k, v in implicitparams.items():
                if a in v:
                    #print(f'parent = {k}')
                    if self.allparams[k]['default'] == 'no':
                        defaults[k] = 'yes'
                    elif self.allparams[k]['default']  == 'yes':
                        defaults[k] = 'no'
                    #print(k, a,  defaults[k])
                    break

        # Merge argsdic onto parsdic. Those values set in command line, 
        # from argsdic, will overwrite the defaults obtained from the par file, 
        # from parsdic. The resulting dictionary, self.iparsdic, is what we will 
        # pass to method runtask.

        self.iparsdic = {**defaults, **argsdic}


    def runtask(self):
        if self.Exit:
            return self.Exit
        r = RunTask(self.taskname, self.iparsdic,self.logFile)
        r.run()

    def printHelp(self):
        self.paramXmlInfo.printHelp()


# __init__.py for pysas
#
#

from . import sastask
from . import parser
from . import param
from . import error
from . import runtask
from . import configutils
from . import init_sas
from . import sasutils
from . import config_pysas
from .version import *
from .odfcontrol import odfcontrol

__version__ = f'pysas - (pysas-{VERSION}) [{SAS_RELEASE}-{SAS_AKA}]'

## For initialization of pySAS from configuration file.
from .configutils import sas_cfg
from .init_sas import initializesas
import os
sas_dir_config     = sas_cfg.get("sas", "sas_dir")
sas_ccfpath_config = sas_cfg.get("sas", "sas_ccfpath")
sas_initialize = True

# Check if SAS_DIR, SAS_PATH, and SAS_CCFPATH are already set.
sas_dir = os.environ.get('SAS_DIR')
sas_path = os.environ.get('SAS_PATH')
sas_ccfpath = os.environ.get('SAS_CCFPATH')
if sas_dir and sas_ccfpath and sas_path:
    if os.path.exists(sas_dir) and os.path.exists(sas_ccfpath) and os.path.exists(sas_path):
        # SAS_DIR, SAS_PATH, and SAS_CCFPATH enviroment variables already set.
        # SAS already initialized.
        sas_initialize = False
        if sas_ccfpath == '/home/idies/workspace/headata/FTP/caldb/data/xmm/ccf':
            sas_cfg['DEFAULT']['on_sci_server'] = 'True'

# Checks if defaults from config file work. Initializes SAS if needed.
if sas_cfg['DEFAULT']['on_sci_server'] == 'False' and sas_initialize:
    if os.path.exists(sas_dir_config) and os.path.exists(sas_ccfpath_config):
        initializesas(sas_dir_config, sas_ccfpath_config)
    elif sas_dir_config != '/does/not/exist' and sas_ccfpath_config != '/does/not/exist':
        print('There is a problem with either SAS_DIR or SAS_CCFPATH in the config file.')
        print('Please set manually to initialize SAS.')
        print('sas_dir....: {}'.format(sas_dir_config))
        print('sas_ccfpath: {}'.format(sas_ccfpath_config))

# Get rid of temporary variables to prevent possible conflicts.
del sas_dir_config, sas_ccfpath_config, sas_initialize, sas_dir, sas_path, sas_ccfpath


# ESA (C) 2000-2021
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# pyeslewsearch.py

import os
import sys
import time
import pysas.pyutils.pyutils as pyutils
from pysas.logger import TaskLogger as TL
from astropy.io import fits
from astropy.io import ascii
from astropy.table import Table, unique, QTable
from astropy import units as u
import numpy as np
import re
import glob


logger = TL('pyeslewsearch')

from .version import VERSION, SAS_RELEASE, SAS_AKA

__version__ = f'pyeslewsearch (pyeslewchain-{VERSION}) [{SAS_RELEASE}-{SAS_AKA}]'


def pyeslewsearch():
    """
    Main program of the task. The function will be called when from the
    command line.
    """

    t_start = time.time()

    coldesc = ['DSSNAME', 'SRCNAME', 'XIMNAME']

    expmaps = []

    for i in os.listdir(os.getcwd()):
        if 'EXPMAP8' in i:
            expmaps.append(i)

    if len(expmaps) == 0:
        print('Empty exposure map list.')
        sys.exit(0)

    expmaps = sorted(expmaps)

    for expmap in expmaps:
        for band_code in ('6', '7', '8'):
            expin = expmap[0:23] + band_code + expmap[24:30]
            imin = expmap[0:17] + 'IMAGE_' + band_code + expmap[24:30]
            srclst = expmap[0:17] + 'SRCLST' + band_code + expmap[24:30]

            ecf = get_ECF(band_code)
            n_events = num_events(imin)

            if n_events > 0:
                eslewsearch_guts(imin, expin, srclst, ecf)
            else:
                print('Empty image ({0}).'.format(imin))
    
    b6list = expmaps[0][0:17] + 'OMSRLI6000.FIT'
    b7list = expmaps[0][0:17] + 'OMSRLI7000.FIT'
    b8list = expmaps[0][0:17] + 'OMSRLI8000.FIT'
   
    srclst6_files = glob.glob('*SRCLST6*')

    if len(srclst6_files) != 0:
        logger.log('info', 'Working on source list files for band 6.')

        status = pyutils.merge_fits(srclst6_files, [1], b6list, columns = None, new_ext = None)
        if status:
            logger.log("error","Failed to merge band 6 lists")

    srclst7_files = glob.glob('*SRCLST7*')
    
    if len(srclst7_files) != 0:
        status = pyutils.merge_fits(srclst7_files, [1], b7list, columns = None, new_ext = None)
        if status:
            logger.log("error","Failed to merge band 7 lists")

    srclst8_files = glob.glob('*SRCLST8*')

    if len(srclst8_files) != 0:
        status = pyutils.merge_fits(srclst8_files, [1], b8list, columns = None, new_ext = None)
        if status:
            logger.log("error","Failed to merge band 8 lists")

    logger.log('info', 'Sizes of each band: B6: {}, B7: {} B8: {}.'.format(len(srclst6_files), len(srclst7_files), len(srclst8_files)))
    # cleaning up...
    logger.log('info', 'Cleaning up...')
    srclistrm = glob.glob('*SRCLST*')

    for file_rm in srclistrm:
        os.remove(file_rm)

    # Append the different band lists into a single list (no source merging)
    create_single_list(b6list, b7list, b8list)
    
    del_list = glob.glob('box*.ds')
    if os.path.isfile('detmask.fits'):
        del_list.append('detmask.fits')
    if os.path.isfile('splinemap.ds'):
        del_list.append('splinemap.ds')

    for f in del_list:
        os.remove(f)

    t_stop = time.time()

    logger.log('info', 'eslewsearch completed in {0} seconds.'.format(round(t_stop - t_start, 2)))


def add_name_cols(mlist, image):
    """
    Adds a column into the source list.

    Args:
        mlist: the original list.
        image: the image file.
    """

    logger.log('debug', 'Running add_name_cols with source list {0} and image {1}.'.format(mlist, image))
    
    coord_string = make_coord_strings(mlist)

    fstrings = make_file_string(image, coord_string)
    # Write strings to data file

    temp_col_fits = 'temp_col_data.fits'

    # Convert name strings into FITS format

    logger.log('debug', 'Running create_fits_from_data.')
    coldesc = ['DSSNAME', 'SRCNAME', 'XIMNAME']
    
    coldata = fstrings
    coldata = np.array(coldata).T.tolist()
    status = pyutils.create_fits_from_data(coldata, coldesc, temp_col_fits)
    
    if status:
        logger.log('error', 'Error while running create_fits_from_data')
    
    status = pyutils.add_column_to_fits(mlist, 1, coldesc, coldata)
    
    if status:
        logger.log('error', 'Error while running add_column.')
        sys.exit(0)
   

def make_file_string(image, coords):
    """
    Creates a string containing basic information from the file.

    Args: 
        image: the name of the image file.
        coords: the coordinates corresponding to the file.

    Output:
        fstring: the formatted string.
    """

    logger.log('debug', 'Running make_file_string with image {0}.'.format(image))

    obsid = image[1:11]
    rev = image[2:6]
    fstrings = []
    
 # Loop over each coordinate string and create a string for import
    for coord in coords:
        dsstring = "ds" + obsid + "_" + coord[0] + coord[1] + ".ds"
        srcname = "xs" + rev + "_" + obsid + "_" + coord[0] + coord[1] + ""
        fstring = (dsstring, srcname, "" + image + "")
        fstrings.append(fstring)
    
    return fstrings


def convert_name(srcin):
    """
    Adds the individual band list into one list.

    Args:
        srcin: the information from the source string.

    Output:
        srcout: the name for the source.
    """

    srcin = srcin.replace("'", '')
    srcout = "XMMSL_J" + srcin[18:].replace(':', '')
    return srcout


def create_single_list(b6list, b7list, b8list):
    """
    Creates a single list containing information from each energy bband.

    Args:
        b6list: the list of the sources in the 6 band.
        b7list: the list of the sources in the 7 band.
        b8list: the list of the sources in the 8 band.
    """

    logger.log('debug', 'Running create_single_list...')

    b6temp = 'temp_b6.dat'
    b7temp = 'temp_b7.dat'
    b8temp = 'temp_b8.dat'
    tempmerge = 'temp_merged.dat'
    mergedlist = b8list[0:19] + 'SSLI0000.FIT'

    # Combine into a single text file with separate columns for each band

    if os.path.isfile(b6list):
        # Dump files into ascii
        with fits.open(b6list) as ff6:
            t = Table(ff6[1].data) 
            t = t['SRCNAME', 'RA', 'DEC', 'RADEC_ERR', 'SCTS', 'SCTS_ERR', 'EXT', 'EXT_ERR', 'DET_ML', 'BG_MAP', 'RATE', 'RATE_ERR','FLUX', 'FLUX_ERR',
            'XIMNAME', 'ID_INST']
            t.write(b6temp, format = 'ascii')

        logger.log('info', 'Working with B6...')

        with open(b6temp) as b6_tf:
            rows_b6 = []

            for line in b6_tf.readlines():
                aux_list = line.split(" ")
                # Add useful records (non-titles and with inst!=0) to output text file
                ii = len(aux_list)
                
                if ii > 0:
                    #for k in range(0, ii):
                        #try:
                            #aux_list[k] = str(np.float32(aux_list[k]))
                        #except:
                            #continue
                    if 'xs' in aux_list[0] and aux_list[15] != 0:
                        # Convert source name to XMMSL format
                        cname = convert_name(aux_list[0])
                        # Write output record\
                        row = [cname, aux_list[1], aux_list[2], aux_list[3], aux_list[4], aux_list[5], np.nan, np.nan, np.nan, np.nan,\
                            aux_list[6], aux_list[7], np.nan, np.nan, np.nan, np.nan, aux_list[8], \
                            np.nan, np.nan, aux_list[9], np.nan, np.nan, aux_list[10], aux_list[11], \
                            np.nan, np.nan, np.nan, np.nan, aux_list[12], aux_list[13], \
                            np.nan, np.nan, np.nan, np.nan, aux_list[14]]

                        rows_b6.append(row)

    else:
        logger.log('error', 'Could not open the B6 list.')

    logger.log('info', 'Working with B7...')
    
    if os.path.isfile(b7list):
        with fits.open(b7list) as ff:
            t = Table(ff[1].data)
            t = t['SRCNAME', 'RA', 'DEC', 'RADEC_ERR', 'SCTS', 'SCTS_ERR', 'EXT' , 'EXT_ERR' ,'DET_ML', 'BG_MAP', 'RATE', 'RATE_ERR','FLUX', 'FLUX_ERR', 'XIMNAME', 'ID_INST']
            ascii.write(t, b7temp, overwrite = True)

        with open(b7temp) as b7_tf:
            rows_b7 = []
            for line in b7_tf.readlines():
                aux_list = line.split(' ')
                # Add useful records to output text file
                ii = len(aux_list)
                if ii > 0:
                    #for k in range(0, ii):
                        #try:
                            #aux_list[k] = str(np.float32(aux_list[k]))
                        #except:
                            #continue
                    if 'xs' in aux_list[0] and aux_list[15] != 0:
                        # Convert source name to XMMSL format
                        cname = convert_name(aux_list[0])
                        # Write output record
                        row = [cname, aux_list[1], aux_list[2], aux_list[3], np.nan, np.nan, aux_list[4], aux_list[5], np.nan, np.nan, \
                            np.nan, np.nan, aux_list[6], aux_list[7], np.nan, np.nan, np.nan, aux_list[8], \
                            np.nan, np.nan, aux_list[9], np.nan, np.nan, np.nan, aux_list[10], aux_list[11], \
                            np.nan, np.nan, np.nan, np.nan, aux_list[12], aux_list[13], \
                            np.nan, np.nan, aux_list[14]]

                        rows_b7.append(row)


    else:
        logger.log('error', 'Could not open the B7 list.')

    logger.log('info', 'Working with B8...')

    if os.path.isfile(b8list):
        with fits.open(b8list) as ff:
            t = Table(ff[1].data)
            t = t['SRCNAME', 'RA', 'DEC', 'RADEC_ERR', 'SCTS', 'SCTS_ERR', 'EXT' , 'EXT_ERR' ,'DET_ML', 'BG_MAP', 'RATE', 'RATE_ERR','FLUX', 'FLUX_ERR', 'XIMNAME', 'ID_INST']
            ascii.write(t, b8temp, overwrite = True)

        with open(b8temp) as b8_tf:
            rows_b8 = []
            for line in b8_tf.readlines():
                aux_list = line.split(" ")
                # Add useful records to output text file
                ii = len(aux_list)
                if ii > 0:
                    #for k in range(0, ii):
                        #try:
                            #aux_list[k] = str(np.float32(aux_list[k]))
                        #except:
                            #continue
                    if 'xs' in aux_list[0] and aux_list[15] != 0:
                        # Convert source name to XMMSL format
                        cname = convert_name(aux_list[0])
                        # Write output record
                        row = [cname, aux_list[1], aux_list[2], aux_list[3], np.nan, np.nan, np.nan, np.nan, aux_list[4], aux_list[5], np.nan, np.nan, \
                            np.nan, np.nan, aux_list[6], aux_list[7], np.nan, np.nan,  aux_list[8], \
                            np.nan, np.nan, aux_list[9], np.nan, np.nan, np.nan, np.nan, aux_list[10], aux_list[11], \
                            np.nan, np.nan, np.nan, np.nan, aux_list[12], aux_list[13], aux_list[14]]

                        rows_b8.append(row)

    else:
        logger.log('error', 'Could not open the B8 list.')

    logger.log('info', 'Write the merged text file into a FITS file.')
    
    labels = make_final_cols()

    logger.log('info', 'Joining band files...')

    start_b6_data = np.array(rows_b6).T.tolist()
    status = pyutils.create_fits_from_data(start_b6_data, labels, 'b6band.temp', 'SRCLIST')
    
    logger.log('debug', 'Creating temp 6')
    start_b7_data = np.array(rows_b7).T.tolist()
    status = pyutils.create_fits_from_data(start_b7_data, labels, 'b7band.temp', 'SRCLIST')
    start_b8_data = np.array(rows_b8).T.tolist()
    status = pyutils.create_fits_from_data(start_b8_data, labels, 'b8band.temp', 'SRCLIST')

    temp_band_files = ('b6band.temp', 'b7band.temp', 'b8band.temp')
    status = pyutils.merge_fits(temp_band_files, [1], mergedlist, columns = None, new_ext = None)

    for f in temp_band_files:
        if os.path.isfile(f):
            os.remove(f)

    #status = pyutils.add_row_to_fits(mergedlist, 'SRCLIST', rows_b7)
    #if status:
    #    logger.log('error', 'Error while running add_rows_to_fits.')
    
    #status = pyutils.add_row_to_fits(mergedlist, 'SRCLIST', rows_b8)
    
    #if status:
    #    logger.log('error', 'Error while running add_row_to_fits.')

    # Finally sort the list on RA, deleting additional entries.

    with fits.open(mergedlist) as ff:
        t = QTable(ff[1].data)
        t = unique(t, 'RA')
        t['RA'] = t['RA'].astype('float')
        t['DEC'] = t['DEC'].astype('float')
        t.sort('RA')

    qt = add_custom_format(t)
    qt = add_units_table(qt)

    try:
        qt.write(mergedlist, format = 'fits', overwrite = True)
    except:
        logger.log('error', 'Could not save the file.')
        sys.exit(0)

    # Change the extension
    with fits.open(mergedlist, 'update') as ef:        
        ef[1].name = 'SRCLIST'
        ef.flush()
   

    add_comments(mergedlist)
    add_units_band(b6list)
    add_units_band(b7list)
    add_units_band(b8list)

    # Clean up but do not throw error if no files to clean
    if os.path.isfile(b6temp):
        os.remove(b6temp)
    if os.path.isfile(b7temp):
        os.remove(b7temp)
    if os.path.isfile(b8temp):
        os.remove(b8temp)
    if os.path.isfile('temp_col_data.fits'):
        os.remove('temp_col_data.fits')


def add_custom_format(t):
    """
    Changes the type of the values for the input data.

    Args:
        t: the input table.
    Output:
        t (corrected).
    """

    aux_srcname = np.chararray(len(t['SRCNAME']), itemsize = 40, unicode = False)
    aux_srcname[:] = t['SRCNAME']
    t['SRCNAME'] = aux_srcname
    t['RA'] = np.array(t['RA'], dtype = 'float64')
    t['DEC'] =  np.array(t['DEC'], dtype = 'float64')
    t['RADEC_ERR'] = np.array(t['RADEC_ERR'], dtype =  'float64')
    t['SCTS_B6'] = np.array(t['SCTS_B6'], dtype = 'float32')
    t['SCTS_B6_ERR'] = np.array(t['SCTS_B6_ERR'], dtype = 'float32')
    t['SCTS_B7'] = np.array(t['SCTS_B7'],dtype = 'float32')
    t['SCTS_B7_ERR'] = np.array(t['SCTS_B7_ERR'], dtype = 'float32')
    t['SCTS_B8'] = np.array(t['SCTS_B8'], dtype = 'float32')
    t['SCTS_B8_ERR'] = np.array(t['SCTS_B8_ERR'], dtype = 'float32')
    t['EXT_B6'] = np.array(t['EXT_B6'], dtype = 'float32')
    t['EXT_B6_ERR'] = np.array(t['EXT_B6_ERR'], dtype = 'float32')
    t['EXT_B7'] = np.array(t['EXT_B7'], dtype = 'float32')
    t['EXT_B7_ERR'] = np.array(t['EXT_B7_ERR'], dtype = 'float32')
    t['EXT_B8'] = np.array(t['EXT_B8'], dtype = 'float32')
    t['EXT_B8_ERR'] = np.array(t['EXT_B8_ERR'], dtype = 'float32')
    t['DET_ML_B6'] = np.array(t['DET_ML_B6'], dtype = 'float32')
    t['DET_ML_B7'] = np.array(t['DET_ML_B7'], dtype = 'float32')
    t['DET_ML_B8'] = np.array(t['DET_ML_B8'], dtype = 'float32')
    t['BG_MAP_B6'] = np.array(t['BG_MAP_B6'], dtype = 'float32')
    t['BG_MAP_B7'] = np.array(t['BG_MAP_B7'], dtype = 'float32')
    t['BG_MAP_B8'] = np.array(t['BG_MAP_B8'], dtype = 'float32')
    t['RATE_B6'] = np.array(t['RATE_B6'], dtype = 'float32')
    t['RATE_B6_ERR'] = np.array(t['RATE_B6_ERR'], dtype = 'float32')
    t['RATE_B7'] = np.array(t['RATE_B7'], dtype = 'float32')
    t['RATE_B7_ERR'] = np.array(t['RATE_B7_ERR'], dtype = 'float32')    
    t['RATE_B8'] = np.array(t['RATE_B8'], dtype = 'float32')
    t['RATE_B8_ERR'] = np.array(t['RATE_B8_ERR'], dtype = 'float32')
    t['FLUX_B6'] = np.array(t['FLUX_B6'], dtype = 'float32')
    t['FLUX_B6_ERR'] = np.array(t['FLUX_B6_ERR'], dtype = 'float32') 
    t['FLUX_B7'] = np.array(t['FLUX_B7'], dtype = 'float32')
    t['FLUX_B7_ERR'] = np.array(t['FLUX_B7_ERR'], dtype = 'float32')
    t['FLUX_B8'] = np.array(t['FLUX_B8'], dtype = 'float32')
    t['FLUX_B8_ERR'] = np.array(t['FLUX_B8_ERR'], dtype = 'float32')
    aux_ximname = np.chararray(len(t['XIMNAME']), itemsize = 40, unicode = False)
    aux_ximname[:] = t['XIMNAME']
    t['XIMNAME'] = aux_ximname

    return t


def add_units_band(blist):
    """
    Moves the input table into a Quantity table object and gets the proper units for each column. Applied to band fits files.

    Args:
        blist: the FITS file of the band.

    Output:
        1 when finished.
    """

    logger.log('debug', 'Running add_units_band with {}.'.format(blist))

    with fits.open(blist) as bl:
        qt = QTable(bl[1].data)

    qt = add_custom_format_band(qt)

    try:
        qt['SCTS'].unit = u.ct
        qt['SCTS_ERR'].unit = u.ct
        qt['X_IMA'].unit = u.pix
        qt['X_IMA_ERR'].unit = u.pix
        qt['Y_IMA'].unit = u.pix
        qt['Y_IMA_ERR'].unit = u.pix
        qt['EXT'].unit = u.pix
        qt['EXT_ERR'].unit = u.pix
        qt['BG_MAP'].unit = u.ct / u.pix
        qt['EXP_MAP'].unit = u.s
        qt['FLUX'].unit = u.erg / u.s / (u.cm * u.cm) 
        qt['FLUX_ERR'].unit = u.erg / u.s / (u.cm * u.cm)
        qt['RATE'].unit = u.ct / u.s
        qt['RATE_ERR'].unit = u.ct / u.s
        qt['RA'].unit = u.deg
        qt['DEC'].unit = u.deg
        qt['RADEC_ERR'].unit = u.arcsec
        qt['LII'].unit = u.deg
        qt['BII'].unit = u.deg
        qt['RAWX'].unit = u.pix
        qt['RAWY'].unit = u.pix
        qt['OFFAX'].unit = u.arcmin
        qt['CUTRAD'].unit = u.pix
        qt['DIST_NN'].unit = u.arcsec
    except:
        logger.log('warning', 'Could not fit all the data in {}.'.format(blist))
    

    qt.write(blist, format = 'fits', overwrite = True)

    return 1


def add_custom_format_band(t):
    """
    Corrects the format of the input data table.

    Args:
        t: the data table for bands B6, B7 and B8.

    Output:
        t (corrected)
    """

    logger.log('debug', 'Running add_custom_format...')
    
    t['ML_ID_SRC'] = np.array(t['ML_ID_SRC'], dtype = 'int32')
    t['BOX_ID_SRC'] = np.array(t['BOX_ID_SRC'], dtype = 'int32')
    t['ID_INST'] = np.array(t['ID_INST'], dtype = 'int32')
    t['ID_BAND'] = np.array(t['ID_BAND'], dtype = 'int32')
    t['ID_CLUSTER'] = np.array(t['ID_CLUSTER'], dtype = 'int32')
    t['SCTS'] = np.array(t['SCTS'], dtype = 'float32')
    t['SCTS_ERR'] = np.array(t['SCTS_ERR'], dtype = 'float32')
    t['X_IMA'] = np.array(t['X_IMA'], dtype = 'float32')
    t['X_IMA_ERR'] = np.array(t['X_IMA_ERR'], dtype = 'float32')
    t['Y_IMA'] = np.array(t['Y_IMA'], dtype = 'float32')
    t['Y_IMA_ERR'] = np.array(t['Y_IMA_ERR'], dtype = 'float32')
    t['EXT'] = np.array(t['EXT'], dtype = 'float32')
    t['EXT_ERR'] = np.array(t['EXT_ERR'], dtype = 'float32')
    t['DET_ML'] = np.array(t['DET_ML'], dtype = 'float32')
    t['EXT_ML'] = np.array(t['EXT_ML'], dtype = 'float32')
    t['BG_MAP'] = np.array(t['BG_MAP'], dtype = 'float32')
    t['EXP_MAP'] = np.array(t['EXP_MAP'], dtype = 'float32')
    t['FLUX'] = np.array(t['FLUX'], dtype = 'float32')
    t['FLUX_ERR'] = np.array(t['FLUX_ERR'], dtype = 'float32')
    t['RATE'] = np.array(t['RATE'], dtype = 'float32')
    t['RATE_ERR'] = np.array(t['RATE_ERR'], dtype = 'float32')
    t['RA'] = np.array(t['RA'], dtype = 'float64')
    t['DEC'] = np.array(t['DEC'], dtype = 'float64')
    t['RADEC_ERR'] = np.array(t['RADEC_ERR'], dtype = 'float32')
    t['LII'] = np.array(t['LII'], dtype = 'float64')
    t['BII'] = np.array(t['BII'], dtype = 'float64')
    t['RAWX'] = np.array(t['RAWX'], dtype = 'int32')
    t['RAWY'] = np.array(t['RAWY'], dtype = 'int32')
    t['OFFAX'] = np.array(t['OFFAX'], dtype = 'float32')
    t['CCDNR'] = np.array(t['CCDNR'] , dtype = 'int32')
    t['HR1'] = np.array(t['HR1'], dtype = 'float32')
    t['HR1_ERR'] = np.array(t['HR1_ERR'], dtype = 'float32')
    t['HR2'] = np.array(t['HR2'], dtype = 'float32')
    t['HR2_ERR'] = np.array(t['HR2_ERR'], dtype = 'float32')
    t['HR3'] = np.array(t['HR3'], dtype = 'float32')
    t['HR3_ERR'] = np.array(t['HR3_ERR'], dtype = 'float32')
    t['HR4'] = np.array(t['HR4'], dtype =  'float32')
    t['HR4_ERR'] = np.array(t['HR4_ERR'], dtype = 'float32')
    t['CUTRAD'] = np.array(t['CUTRAD'], dtype = 'float32')
    t['MASKFRAC'] = np.array(t['MASKFRAC'], dtype = 'float32')
    t['EEF'] = np.array(t['EEF'], dtype = 'float32')
    t['VIGNETTING'] = np.array(t['VIGNETTING'], dtype = 'float32')
    t['ONTIME'] = np.array(t['ONTIME'], dtype = 'float32')
    t['PILEUP'] = np.array(t['PILEUP'], dtype = 'float32')
    t['DIST_NN'] = np.array(t['DIST_NN'], dtype = 'float32')
    aux_flag = np.chararray(len(t['FLAG']), itemsize = 12, unicode = False)
    aux_flag[:] = t['FLAG']
    t['FLAG'] = aux_flag
    aux_DSSNAME = np.chararray(len(t['DSSNAME']), itemsize = 40, unicode = False)
    aux_DSSNAME[:] = t['DSSNAME']
    t['DSSNAME'] = aux_DSSNAME
    aux_srcname = np.chararray(len(t['SRCNAME']), itemsize = 40, unicode = False)
    aux_srcname[:] = t['SRCNAME']
    t['SRCNAME'] = aux_srcname
    aux_ximname = np.chararray(len(t['XIMNAME']), itemsize = 40, unicode = False)
    aux_ximname[:] = t['XIMNAME']
    t['XIMNAME'] = aux_ximname
    #DSSNAME   str40
    #SRCNAME   str40
    #XIMNAME   str40

    return t


def add_units_table(qt):
    """
    Moves the input table into a Quantity table object and gets the proper units for each column.

    Args:
        qt: the input table without units.

    Output:
        qt: the Quantity Table with proper units.
    """
    
    logger.log('debug', 'Adding units to main table...')
    
    try:
        qt['RA'] = qt['RA'] * u.deg
        qt['DEC'].unit = u.deg
        qt['RADEC_ERR'].unit = u.arcsec
        qt['EXT_B6'].unit = u.pix
        qt['EXT_B6_ERR'].unit = u.pix
        qt['EXT_B7'].unit = u.pix
        qt['EXT_B7_ERR'].unit = u.pix
        qt['EXT_B8'].unit = u.pix
        qt['EXT_B8_ERR'].unit = u.pix
        qt['BG_MAP_B6'].unit = u.ct / u.pix
        qt['BG_MAP_B7'].unit = u.ct / u.pix
        qt['BG_MAP_B8'].unit = u.ct / u.pix
        qt['RATE_B6'].unit = u.ct / u.s
        qt['RATE_B6_ERR'].unit = u.ct / u.s
        qt['RATE_B7'].unit = u.ct / u.s
        qt['RATE_B7_ERR'].unit = u.ct / u.s
        qt['RATE_B8'].unit = u.ct / u.s
        qt['RATE_B8_ERR'].unit = u.ct / u.s
        qt['FLUX_B6'].unit = u.erg / u.s / (u.cm * u.cm)
        qt['FLUX_B6_ERR'].unit = u.erg / u.s / (u.cm * u.cm)
        qt['FLUX_B7'].unit = u.erg / u.s / (u.cm * u.cm)
        qt['FLUX_B7_ERR'].unit = u.erg / u.s / (u.cm * u.cm)
        qt['FLUX_B8'].unit = u.erg / u.s / (u.cm * u.cm)
        qt['FLUX_B8_ERR'].unit = u.erg / u.s / (u.cm * u.cm)
    except:
        logger.log('warning', 'An error ocurred while adding the units to the final source list. It may be incomplete.')

    return qt


def num_events(image):
    """
    Gets the total number of events in a given image.

    Args:
        image: the path or FITS image to be evaluated.
    Output:
        sum: the total sum returned by imgstat.
    """

    stat_image = pyutils.imgstat(image, 0, None)

    return stat_image[4]


def make_coord_strings(mlist):
    """
    Reads coordinates and converts them into HMS,dms.

    Args:
        mlist: list to check coordinates.

    Output:
        new_coords: a list of the tuples of the new coordinates.
    """

    new_coords = []
    coords = pyutils.pydump(mlist, 1, ['RA', 'DEC'])

    for row in coords:
        ra_o = float(row[0])
        dec_o = float(row[1])
        values_coords = (convert_RA(ra_o), convert_DEC(dec_o))
        new_coords.append(values_coords)
    return new_coords


def make_final_cols():
    """
    Returns the main labels to use in the column to be formatted.

    Output:
        labels: the labels to be used.
    """
    
    labels = ['SRCNAME',
 'RA',
 'DEC',
 'RADEC_ERR',
 'SCTS_B6',
 'SCTS_B6_ERR',
 'SCTS_B7',
 'SCTS_B7_ERR',
 'SCTS_B8',
 'SCTS_B8_ERR',
 'EXT_B6',
 'EXT_B6_ERR',
 'EXT_B7',
 'EXT_B7_ERR',
 'EXT_B8',
 'EXT_B8_ERR',
 'DET_ML_B6',
 'DET_ML_B7',
 'DET_ML_B8',
 'BG_MAP_B6',
 'BG_MAP_B7',
 'BG_MAP_B8',
 'RATE_B6',
 'RATE_B6_ERR',
 'RATE_B7',
 'RATE_B7_ERR',
 'RATE_B8',
 'RATE_B8_ERR',
 'FLUX_B6',
 'FLUX_B6_ERR',
 'FLUX_B7',
 'FLUX_B7_ERR',
 'FLUX_B8',
 'FLUX_B8_ERR',
 'XIMNAME']

    return labels


def eslewsearch_guts(image, expmap, mlist, ecf):
    """
    Runs several SAS' tasks to create additional files. Will quit if the aforementioned taasks run into any error (check the log file).

    Args:
        image: the image to manipulate.
        expmap: the exposure map for the image.
        mlist:
        ecf: the energy correction factor.
    """

    logger.log('debug', 'Running eslewsearch_guts with {0}, {1}, {2}.'.format(image, expmap, mlist))

    # Set constants and thresholds
    pimin = 1499 # USe 1.5 keV for the PSF in all energy-band images!!
    pimax = 1501
    eboxlthr = 8
    eboxmthr = 10
    emlthr = 8
    mask = 'detmask.fits'
    boxlist_l = 'boxlist_l.ds'
    boxlist_m = 'boxlist_m.ds'
    bckgnd = 'splinemap.ds'

    # Run emask
    status_code = os.system("emask expimageset={0} detmaskset={1} threshold1=1e-5 threshold2=0.5".format(expmap, mask))
    
    if status_code:
        logger.log('error', 'Could not run emask.')
        sys.exit(0)

    # Run eboxdetect (local mode)
    status_code = os.system("eboxdetect boxlistset={0} boxsize=5 expimagesets={1} imagesets={2} likemin={3} nruns=3 pimax={4} pimin={5} usemap=no usematchedfilter=no withdetmask=yes detmasksets={6} withexpimage=yes withoffsets=no obsmode='slew'".format(boxlist_l, expmap, image, eboxlthr, pimax, pimin, mask))
    
    if status_code:
        logger.log("error","Failure in eboxdetect (local) on {0}".format(image))
        sys.exit(0)

    # Create background map
    status_code = os.system("esplinemap bkgimageset={0} boxlistset={1} withcheese=no withdetmask=yes detmaskset={2} excesssigma=3 idband=1 imageset={3} mlmin=1 nfitrun=3 nsplinenodes=10 scut=0.005 withexpimage=no".format(bckgnd, boxlist_l, mask, image))
    if status_code:
        logger.log("error","Failure in esplinemap on {0}.".format(image))
        sys.exit(0)

    # Run eboxdetect (map mode)
    status_code = os.system("eboxdetect bkgimagesets={0} boxlistset={1} boxsize=5 withexpimage=yes expimagesets={2} imagesets={3} likemin={4} nruns=3 pimax={5} pimin={6} usemap=yes usematchedfilter=no withdetmask=yes detmasksets={7} withoffsets=no obsmode='slew'".format(bckgnd, boxlist_m, expmap, image, eboxmthr, pimax, pimin, mask))
    
    if status_code:
        logger.log("error","Failure in eboxdetect (map) on {0}".format(image));
        sys.exit(0)

    # Run emldetect
    status_code = os.system("emldetect bkgimagesets={0} boxlistset={1} determineerrors=yes dmlextmin=2 ecf={2} ecut=36 withexpimage=yes expimagesets={3} fitcounts=yes fitextent=yes fitposition=yes imagesets={4} mllistset={5} mlmin={6} pimax={7} pimin={8} scut=0.9 usecalpsf=yes useevents=no withhotpixelfilter=no withoffsets=no withxidband=no nmaxfit=1 nmulsou=1 psfmodel='slew'".format(bckgnd, boxlist_m, ecf, expmap, image, mlist, emlthr, pimax, pimin))
    
    if status_code:
        logging.log("error","Failure in emldetect on {0}".format(image))
        sys.exit(0)

    if os.path.isfile(mlist):
        add_name_cols(mlist, image)


def convert_RA(ra):
    """
    Change the format of the RA parameter and returns it as a formatted string.

    Args:
        ra: the numerical value of RA.

    Output:
        hms: the formatted RA string.
    """

    hrs = int(ra / 15)
    mins = int((ra - hrs * 15.0) * 60.0 / 15.0)
    secs = (ra - (hrs * 15.0 + mins * 15.0 / 60.0)) * 3600.0 / 15.0

    if secs > 59.95:
        secs = 0
        mins = mins + 1

    if mins == 60:
        mins = 0
        hrs = hrs + 1

    if hrs == 24:
        hrs = 0

    #hms = '{0}:{1}:{2}'.format(str(round(hrs, 2)).zfill(2), str(round(mins, 2)).zfill(2), round(secs, 4))

    hms = '{0}:{1}:{2}'.format(str(round(hrs, 2)).zfill(2), str(round(mins, 2)), round(secs, 1))

    return hms


def convert_DEC(dec):
    """
    Converts the dec into a formatted string.

    Args:
        dec: the dec value as a real number.

    Output:
        degstring:
    """

    deg = int(dec)
    mins = int((dec - deg) * 60.0)
    secs = int(abs(dec - (deg + mins / 60.0)) * 3600.0 + 0.5)
    amins = abs(mins)

    # Sign
    sign = "+"

    if (dec < 0.0):
        sign = "-"

    # Sort out ranges
    if (secs > 59.95):
        secs = 0.0
        amins = amins + 1

    if amins == 60:
        amins = 0.0
        if dec < 0.0:
            deg = deg - 1
        else:
            deg = deg + 1

    degstring = '{0}{1}:{2}:{3}'.format(sign, str(round(abs(deg), 2)).zfill(2), str(round(amins, 2)).zfill(2), str(round(secs, 2)).zfill(2))

    return degstring;


def get_ECF(band):
    """
    Returns the energy conversion factor for the corresponding energy band.

    Args:
        band: the energy band (6, 7, 8).

    Output:
        ecf: the energy conversion factor (as a real).
    """

    band = int(band)

    if band == 6:
        ecf = 1.436 # soft energy band (0.2 - 2 keV)
    elif band == 7:
        ecf = 9.144 # hard energy band (2 - 12 keV)
    elif band == 8:
        ecf = 3.159 # total energy band (0.2 - 12 keV)
    else:
        logger.log('error', 'The input energy band is not supported ({0})'.format(band))
        sys.exit(0)

    ecf = 10 / ecf # for EMLDETECT

    return ecf


def add_comments(mlist):
    """
    Adds the comment info to the first header of the source list.

    Args:
        mlist: the final source list.
    """

    with fits.open(mlist, 'update') as ff:
        for j in ff[1].header:
            field_n = re.findall(r'\d+', j)
            if len(field_n) > 0:
                field_n = field_n[0]
            if ff[1].header[j] == 'RA':
                ff[1].header[j] = ('RA', 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'DEC':
                ff[1].header[j] = ("DEC", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'RADEC_ERR':
                ff[1].header[j] = ("RADEC_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == "EXT_B6":
                ff[1].header[j] = ("EXT_B6", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'EXT_B6_ERR':
                ff[1].header[j] = ("EXT_B6_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'EXT_B7':
                ff[1].header[j] = ("EXT_B7", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'EXT_B7_ERR':
                ff[1].header[j] = ("EXT_B7_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'EXT_B8':
                ff[1].header[j] = ("EXT_B8", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'EXT_B8_ERR':
                ff[1].header[j] = ("EXT_B8_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'BG_MAP_B6':
                ff[1].header[j] = ("BG_MAP_B6", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'BG_MAP_B7':
                ff[1].header[j] = ("BG_MAP_B7", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'BG_MAP_B8':
                ff[1].header[j] = ("BG_MAP_B8", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'RATE_B6':
                ff[1].header[j] = ("RATE_B6", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'RATE_B6_ERR':
                ff[1].header[j] = ("RATE_B6_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'RATE_B7':
                ff[1].header[j] = ("RATE_B7", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'RATE_B7_ERR':
                ff[1].header[j] = ("RATE_B7_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'RATE_B8':
                ff[1].header[j] = ("RATE_B8", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'RATE_B8_ERR':
                ff[1].header[j] = ("RATE_B8_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'FLUX_B6':
                ff[1].header[j] = ("FLUX_B6", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'FLUX_B6_ERR':
                ff[1].header[j] = ("FLUX_B6_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'FLUX_B7':
                ff[1].header[j] = ("FLUX_B7", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'FLUX_B7_ERR':
                ff[1].header[j] = ("FLUX_B7_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'FLUX_B8':
                ff[1].header[j] = ("FLUX_B8", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'FLUX_B8_ERR':
                ff[1].header[j] = ("FLUX_B8_ERR", 'label for field {}'.format(field_n))
        ff.flush()


def run(iparsdic):
    print('Executing {}.'.format(__file__))
    pyeslewsearch()




# ESA (C) 2000-2021
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#


# Standard library imports
import logging
import os

# Third party imports

# Local application imports


class TaskLogger:
    """
    TaskLogger creates a logger for Python tasks.
    The name of the logger is the name of the task.
    Such logger is known as saslogger. 
    It includes two handlers, one console handler
    and one file handler. The file handler will put
    all logging messages in a file named <taskname>.log.
    
    This file is by default written in the starting directory
    of the task (os.getcwd()). However, for each task we can 
    set that directory by taking into account the env variable 
    SAS_TASKLOGDIR, if defined. 
    
    By default the logging file is created in  mode -a- (append)
    to allow to use the same file to log several runs of the
    same task. However, the SAS_TASKLOGFMODE environment variable
    can be set to -w- (new file), to change that behaviour.

    File handler has alogging level of DEBUG, so everything
    is logged. However, console handler logging level is 
    controlled by the SAS Verbosity (SAS_VERBOSITY env. variable)
    which is set via the option -V.

    The logger can not be called more than once. If not, 
    messages start appearing duplicated. That is why the
    logging method is separarted from the logger definition.
    """

    def __init__(self, taskname):
        self.taskname = taskname
       
        self.dictlevels = {
                'notset':0,
                'debug': 10,
                'info': 20,
                'warning': 30,
                'error': 40,
                'critical': 50
                }

        self.saslogger = logging.getLogger(self.taskname)
        self.saslogger.setLevel('DEBUG')

        # SAS_TASKLOGDIR allows to set the directory for tyhe logging file
        sastasklogdir = os.getenv('SAS_TASKLOGDIR')
        startdir = os.getcwd()
        #print(f'{self.taskname} was initiated from {startdir}')
       
        if not sastasklogdir:
            #print(f'\nTask logging directory is unset. \nUsing {startdir} instead.')
            sastasklogdir = startdir
        if not os.path.isdir(sastasklogdir):
            #print(f'\n{sastasklogdir} does not exist. \nUsing {startdir} instead.')
            sastasklogdir = startdir

        self.logfile = os.path.join(sastasklogdir, self.taskname + '.log')
        #print(f'\n{self.taskname} log file: {self.logfile}\n')

        # Clear any previous handlers
        if self.saslogger.hasHandlers:
            self.saslogger.handlers.clear()

        # SAS_TASKLOGFMODE allows to set the write mode for the logging file
        # Allowed modes are : w (new file each invocation of logger),
        # a (append to any existing file)A That is the default mode.

        sastasklogfmode = os.getenv('SAS_TASKLOGFMODE')
        if not sastasklogfmode:
            sastasklogfmode = 'a'
        if sastasklogfmode != 'a' and sastasklogfmode != 'w':
            sastasklogfmode = 'a'
        #print(f'\nTask logging file mode = {sastasklogfmode}')

        # File handler (fh). Level fixed to DEBUG
        self.fh = logging.FileHandler(self.logfile, mode=sastasklogfmode)
        self.fh.setLevel('DEBUG')

        # Console handler (ch). Default level set to WARNING
        self.ch = logging.StreamHandler()
        self.ch.setLevel('WARNING')

        # Add handlers to the logger
        self.saslogger.addHandler(self.ch)
        self.saslogger.addHandler(self.fh)

        # Create formatters and add them to handlers
        cformat = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
        fformat = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s', datefmt='%d-%b-%y %H:%M:%S')

        self.ch.setFormatter(cformat)
        self.fh.setFormatter(fformat)


    def __repr__(self):
        return f'{self.__class__.__name__}({self.taskname})'


    def log(self, level, msg):
        """
        A log message will be issued according to the logging level.
        The logging level is fixed for the log file but 
        on the console will depend on SAS_VERBOSITY.
        SAS_VERBOSITY can be set between 1 (min) and 10 (max).
        Correspondence between SAS_VERBOSITY levels and logging levels
        DEBUG      ->   verbosity  = 8, 9, 10
        INFO       ->   verbosity  = 6, 7
        WARNING    ->   verbosity  = 4, 5
        ERROR      ->   verbosity  = 2, 3
        CRITICAL   ->   verbosity  = 1
        """

        v = os.getenv('SAS_VERBOSITY')
        if v == None:
            v = '4'
        verbosity = int(v)

        if   verbosity >= 8 and verbosity <= 10:
            self.ch.setLevel('DEBUG')
        elif verbosity >= 6 and verbosity <= 7:
            self.ch.setLevel('INFO')
        elif verbosity >= 4 and verbosity <= 5:
            self.ch.setLevel('WARNING')
        elif verbosity >= 2 and verbosity <= 3:
            self.ch.setLevel('ERROR')
        elif verbosity == 1:
            self.ch.setLevel('CRITICAL')

        nlevel = self.dictlevels[level]
        self.saslogger.log(nlevel, msg)




# ESA (C) 2000-2020
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# sasver.py

import os
from .version import * 

__version__ = f'sasver (sasver-{VERSION}) [{SAS_RELEASE}-{SAS_AKA}]' 

'''
Shows release and information about the XMM-Newton Science Analysis System (SAS).
'''

def run(args):
    sas_dir = os.environ.get('SAS_DIR')
    if sas_dir == None:
        print('Error: sasver : SAS_DIR is undefined')
        sys.exit()


    sas_path = os.environ.get('SAS_PATH')
    sas_ccfpath = os.environ.get('SAS_CCFPATH')
    sas_ccf = os.environ.get('SAS_CCF')
    sas_odf = os.environ.get('SAS_ODF')
    sas_mmod = os.environ.get('SAS_MEMORY_MODEL')

    print(f'''
    XMM-Newton SAS - release and build information

    SAS release: {SAS_RELEASE}-{SAS_AKA}
    Compiled on: {SAS_COMPILATION_DATE}
    Compiled by: {SAS_COMPILATION_USER}@{SAS_COMPILATION_HOST}
    Platform   : {SAS_PLATFORM}

    SAS-related environment variables set:

    ''')
    print(f'SAS_DIR        = {sas_dir}')
    print(f'SAS_PATH       = {sas_path}')
    print(f'SAS_CCFPATH    = {sas_ccfpath}') if sas_ccfpath != None else None
    print(f'SAS_CCF        = {sas_ccf}') if sas_ccf != None else None
    print(f'SAS_ODF        = {sas_odf}') if sas_odf != None else None




# ESA (C) 2000-2024
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# sasutils.py

"""
sasutils.py

Utility functions specific to SAS or pySAS.
"""

# Standard library imports
import os, sys, subprocess, shutil, glob, tarfile, gzip, time, platform, re

# Third party imports
from astroquery.esa.xmm_newton import XMMNewton

# Local application imports
# from .version import VERSION, SAS_RELEASE, SAS_AKA
from .logger import TaskLogger as TL

# Third party imports

# Local application imports

def download_data(odfid,
                  data_dir,
                  level='ODF',
                  repo='esa',
                  logger=None,
                  encryption_key=None,
                  proprietary=False,
                  credentials_file=None,
                  overwrite=True,
                  PPS_subset=False,
                  instname=None,
                  expflag=None,
                  expno=None,
                  product_type=None,
                  datasubsetno=None,
                  sourceno=None,
                  extension=None,
                  filename=None,
                  **kwargs):
    """
    --Not intended to be used by the end user. Internal use only.--
    --Use odf.download_data() or odf.basic_setup() instead.--

    Downloads, or copies, data from chosen repository. 

    Will silently overwrite any preexisting data files and remove any existing
    pipeline products. Will create directory structure in 'data_dir' for odf.

    Inputs:

        --odfid:          (string): ID of ODF in string format.

        --data_dir:  (string/path): Path to directory where the data will be 
                                    downloaded. Automatically creates directory
                                    data_dir/odfid.
                                    Default: --REQUIRED-- MUST EXIST!.

        --level:          (string): Level of data products to download.
                                    Default: 'ODF'
                                    Can be 'ODF, 'PPS' or 'ALL'.

        --repo:           (string): Which repository to use to download data. 
                                    Default: 'esa'
                                    Can be either
                                        'esa' (data from Europe/ESA) or 
                                        'heasarc' (data from North America/NASA) or
                                        'sciserver' (if user is on sciserver)

        --logger      (TaskLogger): Task logger object.

        --encryption_key: (string): Encryption key for proprietary data, a string 32 
                                    characters long. -OR- path to file containing 
                                    ONLY the encryption key.

        --proprietary    (boolean): Flag for downloading proprietary data from
                                    the XSA at ESA.

        --credentials_file (filename): Path and filename of file containing XSA
                                       username and password. For proprietary data
                                       only. (Optinal, astroquery will ask user 
                                       for username and password if filename
                                       not given.)

        
    """

    if not logger:
        logger = generate_logger(logname=f'download_{odfid}',log_dir=data_dir)

    # Set directories for the observation, odf, and working
    obs_dir = os.path.join(data_dir,odfid)
    odf_dir = os.path.join(obs_dir,'ODF')
    pps_dir = os.path.join(obs_dir,'PPS')
    # work_dir = os.path.join(obs_dir,'work')

    # Checks if obs_dir exists. Removes it if overwrite=True.
    if os.path.exists(obs_dir) and overwrite:
        logger.log('info', f'Removing existing directory {obs_dir} ...')
        print(f'\n\nRemoving existing directory {obs_dir} ...')
        shutil.rmtree(obs_dir)
    
    # Creates subdirectory odfid to move or unpack observation files
    # and makes subdirectories.
    if not os.path.exists(obs_dir):
        logger.log('info', f'Creating observation directory {obs_dir} ...')
        print(f'\nCreating observation directory {obs_dir} ...')
        os.mkdir(obs_dir)

    logger.log('info', 'Requesting odfid = {} from XMM-Newton Science Archive\n'.format(odfid))
    print('Requesting odfid = {} from XMM-Newton Science Archive\n'.format(odfid))

    if level == 'PPS' and PPS_subset:
        PPSfile = generate_PPS_filename(obsid=odfid,instname=instname,
                                        expflag=expflag,expno=expno,
                                        product_type=product_type,
                                        datasubsetno=datasubsetno,
                                        sourceno=sourceno,extension=extension)
        
    if repo == 'esa':
        logger.log('info', f'Changed directory to {obs_dir}')
        os.chdir(obs_dir)
        if level == 'ALL':
            level = ['ODF','PPS']
        else:
            level = [level]
        for levl in level:
            # Download the odfid from ESA, using astroquery
            logger.log('info', f'Downloading {odfid}, level {levl} into {obs_dir}')
            print(f'\nDownloading {odfid}, level {levl} into {obs_dir}. Please wait ...\n')
            if levl == 'PPS':
                # If a filename was provided then convert it into inputs for astroquery.
                if filename:
                    PPS_subset   = True
                    instname     = filename[11:13]
                    expflag      = filename[13]
                    expno        = filename[14:17]
                    product_type = filename[17:23]
                    datasubsetno = filename[23]
                    sourceno     = filename[24:27]
                    extension    = filename[-3:]
                # Take care of the optional inputs.
                if instname:     kwargs['instname']     = instname
                if expflag:      kwargs['expflag']      = expflag
                if expno:        kwargs['expno']        = expno
                if product_type: kwargs['name']         = product_type
                if datasubsetno: kwargs['datasubsetno'] = datasubsetno
                if sourceno:     kwargs['sourceno']     = sourceno
                if extension:    kwargs['extension']    = extension
            XMMNewton.download_data(odfid, level=levl,
                                    prop=proprietary,
                                    credentials_file=credentials_file,
                                    **kwargs)
            
            if levl == 'ODF':    
                os.mkdir(odf_dir)

            if PPS_subset:
                if not os.path.exists(pps_dir): os.mkdir(pps_dir)
                files = glob.glob(obs_dir+'/*.*')
                for file in files:
                    file_name = os.path.basename(file)
                    shutil.copy(file, os.path.join(pps_dir,file_name))
            else:
                # Check that the tar.gz file has been downloaded
                odftar = glob.glob(obs_dir+f'/{odfid}'+'*')[0]
                try:
                    os.path.exists(odftar)
                    logger.log('info', f'{odftar} found.') 
                except FileExistsError:
                    logger.log('error', f'File {odftar} is not present. Not downloaded?')
                    print(f'File {odftar} is not present. Not downloaded?')
                    sys.exit(1)

                tarextension = os.path.splitext(odftar)[1]
                if tarextension == '.gz': tar_mode = 'r:gz'
                elif tarextension == '.tar': tar_mode = 'r'
                else:
                    logger.log('error', f'File {odftar} extension not recognized.')
                    raise Exception(f'File {odftar} extension not recognized.')

                # Untars the odfid.tar.gz file
                logger.log('info', f'Unpacking {odftar} ...')
                print(f'\nUnpacking {odftar} ...\n')

                try:
                    with tarfile.open(odftar,tar_mode) as tar:
                        if levl == 'ODF':
                            tar.extractall(path=odf_dir)
                        elif levl == 'PPS':
                            tar.extractall(path=data_dir)
                            os.rename('pps','PPS')
                    os.remove(odftar)
                    logger.log('info', f'{odftar} extracted successfully!')
                    logger.log('info', f'{odftar} removed')
                except tarfile.ExtractError:
                    logger.log('error', 'tar file extraction failed')
                    raise Exception('tar file extraction failed')
    elif repo == 'heasarc':
        #Download the odfid from HEASARC, using wget
        logger.log('info', f'Changed directory to {data_dir}')
        os.chdir(data_dir)
        if level == 'ALL':
            levl = ''
        else:
            levl = level

        if PPS_subset:
            PPS_subset_note = f' using file pattern {PPSfile}'
            wgetA = f"-A '{PPSfile}'"
        else:
            PPS_subset_note = ''
            wgetA = ''

        if filename != None:
            wgetf = filename
            PPS_subset_note = f', file {wgetf}'
        
        logger.log('info', f'Downloading {odfid}, level {levl}{PPS_subset_note}')
        print(f'\nDownloading {odfid}, level {level}{PPS_subset_note}. Please wait ...\n')
        cmd = f'wget -m -nH -e robots=off --cut-dirs=4 -l 2 -np {wgetA} https://heasarc.gsfc.nasa.gov/FTP/xmm/data/rev0/{odfid}/{levl}/{wgetf}'
        logger.log('info', f'Using the command:\n{cmd}')
        result = subprocess.run(cmd, shell=True)
        if result.returncode != 0:
            print(f'Problem downloading data!')
            logger.log('error', f'File download failed!')
            raise Exception('File download failed!')
        for path, directories, files in os.walk(obs_dir):
            for file in files:
                if 'index.html' in file:
                    os.remove(os.path.join(path,file))
            for direc in directories:
                if '4XMM' in direc:
                    shutil.rmtree(os.path.join(path,direc))
                if level == 'ODF' and direc == 'PPS':
                    shutil.rmtree(os.path.join(path,direc))
                if level == 'PPS' and direc == 'ODF':
                    shutil.rmtree(os.path.join(path,direc))
                if 'om_mosaic' in direc:
                    shutil.rmtree(os.path.join(path,direc))
    elif repo == 'sciserver':
        # Copies data into personal storage space.
        dest_dir = obs_dir
        if level == 'ALL':
            levl = ''
        else:
            levl = level
            dest_dir = os.path.join(dest_dir,levl)
        if levl == 'ODF':    
            os.mkdir(odf_dir)
        elif levl == 'PPS':
            os.mkdir(pps_dir)
        archive_data = f'/home/idies/workspace/headata/FTP/xmm/data/rev0//{odfid}/{levl}'
        if filename: 
            PPS_subset = True
            PPSfile = filename
        if PPS_subset:
            files = glob.glob(archive_data + f'/**/{PPSfile}', recursive=True)
            for file in files:
                file_name = os.path.basename(file)
                logger.log('info', f'Copying file {file_name} from {archive_data} ...')
                print(f'\nCopying file {file_name} from {archive_data} ...')
                shutil.copy(file, os.path.join(pps_dir,file_name))
        else:
            logger.log('info', f'Copying data from {archive_data} ...')
            print(f'\nCopying data from {archive_data} ...')
            shutil.copytree(archive_data,dest_dir,dirs_exist_ok=True)

    # Check if data is encrypted. Decrypt the data.
    encrypted = glob.glob('**/*.gpg', recursive=True)
    if len(encrypted) > 0:
        logger.log('info', f'Encrypted files found! Decrypting files!')

        # Checks for encryption key or file with key.
        # If no encryption key is given then go looking for a file.
        encryption_file = None
        if encryption_key == None:
            encryption_file = glob.glob(os.path.join(data_dir,f'*{odfid}*'))
            if len(encryption_file) == 0:
                encryption_file = glob.glob(os.path.join(data_dir,'*key*'))
            if len(encryption_file) > 1:
                logger.log('error', 'Multiple possible encryption key files. Specify encryption key file.')
                raise Exception('Multiple possible encryption key files.')
            if len(encryption_file) == 0:
                encryption_file = 'None'
            if os.path.isfile(encryption_file[0]):
                logger.log('info', f'File with encryption key found: {encryption_file}')
            else:
                print('File decryption failed. No encryption key found.')
                print(f'Regular file with the encryption key needs to be placed in: {data_dir}')
                logger.log('error', 'File decryption failed. No encryption key found.')
                raise Exception('File decryption failed. No encryption file found.')
        elif os.path.isfile(encryption_key):
            logger.log('info', f'Ecryption key is in file: {encryption_key}')
            encryption_file = encryption_key
        if encryption_file is not None:
            logger.log('info', f'Reading ecryption key from: {encryption_file}')
            with open(encryption_file) as f:
                lines = f.readlines()
                encryption_key = lines[0]
        if encryption_key == None:
            print(f'No encryption key found in {encryption_file}')
            print(f'Regular file with the encryption key needs to be placed in: {data_dir}')
            logger.log('error', 'File decryption failed. No encryption key found.')
            raise Exception('File decryption failed. No encryption key found.')
        
            
        for file in encrypted:
            out_file = file[:-4]
            if os.path.exists(out_file):
                logger.log('info', f'Already decrypted file found: {out_file}')
                print(f'Already decrypted file found: {out_file}')
            else:
                logger.log('info', f'Decrypting {file}')
                cmd = 'echo {0} | gpg --batch -o {1} --passphrase-fd 0 -d {2}'.format(encryption_key,out_file,file)
                result = subprocess.run(cmd, shell=True)
                if result.returncode != 0:
                    print(f'Problem decrypting {file}')
                    logger.log('error', f'File decryption failed, key used {encryption_key}')
                    raise Exception('File decryption failed')
                os.remove(file)
                logger.log('info', f'{file} removed')
    else:
        logger.log('info','No encrypted files found.')

    for file in glob.glob('**/*.gz', recursive=True):
        logger.log('info', f'Unpacking {file} ...')
        print(f'Unpacking {file} ...')
        with gzip.open(f'{file}', 'rb') as f_in:
            out_file = file[:-3]
            with open(out_file, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)
        os.remove(file)
        logger.log('info', f'{file} removed')

    for file in glob.glob('**/*.TAR', recursive=True):
        logger.log('info', f'Unpacking {file} ...')
        print(f'Unpacking {file} ...')
        with tarfile.open(file,"r") as tar:
            tar.extractall(path=odf_dir)
        os.remove(file)
        logger.log('info', f'{file} removed')

    ppssumhtml = 'P' + odfid + 'OBX000SUMMAR0000.HTM'
    ppssumhtmlfull = os.path.join(pps_dir, ppssumhtml)

    if os.path.exists(ppssumhtmlfull):
        ppssumhtmllink = 'file://' + ppssumhtmlfull
        logger.log('info', f'PPS products can be found in {pps_dir}')
        print(f'\nPPS products can be found in {pps_dir}\n\nLink to Observation Summary html: {ppssumhtmllink}')

    return

def generate_logger(logname=None,log_dir=None):
    """
    --Not intended to be used by the end user. Internal use only.--
    """
    if not logname:
        logname = 'general_sas'

    sastasklogdir = os.environ.get('SAS_TASKLOGDIR')

    # Where are we?
    startdir = os.getcwd()

    # Check where the logger should go.
    if log_dir:
        sastasklogdir = log_dir

    if not sastasklogdir:
        sastasklogdir = startdir

    if not os.path.isdir(sastasklogdir):
        sastasklogdir = startdir

    # This will put the log files in data_dir.
    os.environ['SAS_TASKLOGDIR'] = sastasklogdir

    # Create the logger
    logger = TL(logname)

    return logger

def update_calibration_files(repo='NASA'):
    """
    Function to download/update XMM calibration files.
    """

    sas_ccfpath = os.environ.get('SAS_CCFPATH')
    if not sas_ccfpath:
        raise Exception('SAS_CCFPATH not set. Please define it.')

    esa_or_nasa = repo.lower()

    esa = ['esa','e','es','europe']
    nasa = ['nasa','n','na','nas','ns','nsa','us','usa','heasarc','hea']

    if esa_or_nasa in esa:
        cmd = f'rsync -v -a --delete --delete-after --force --include=\'*.CCF\' --exclude=\'*/\' sasdev-xmm.esac.esa.int::XMM_VALID_CCF {sas_ccfpath}'
    elif esa_or_nasa in nasa:
        cmd = f'wget -nH --no-remove-listing -N -np -r --cut-dirs=4 -e robots=off -l 1 -R "index.html*" https://heasarc.gsfc.nasa.gov/FTP/xmm/data/CCF/ -P {sas_ccfpath}'
    print(f'Downloading calibration data using the command:\n{cmd}')
    print('This may take a while.')
    time.sleep(3)
    result = subprocess.run(cmd, shell=True)

    return result

def install_sas(repo='NASA',sas_version='21.0.0'):
    """
    !!WARNING!! EXPERIMENTAL!
    No guarentees this will work.
    Only tested with Ubuntu v. 20 and 22
    """

    tar_file = f'sas_{sas_version}-'

    if repo == 'NASA':
        base_path = f'https://heasarc.gsfc.nasa.gov/FTP/xmm/software/sas/{sas_version}/'
    elif repo == 'ESA':
        base_path = f'https://sasdev-xmm.esac.esa.int/pub/sas/{sas_version}/'
    else:
        raise Exception(f"Repository '{repo}' not recognized. Use either 'NASA' or 'ESA'.")
        

    system = platform.system()
    
    if system == 'Linux':
        base_path = base_path+f'{system}/'
        output = subprocess.run(['lsb_release','-d'],capture_output=True,text=True)
        outlist = output.stdout.split()
        distribution = outlist[1]
        version = outlist[2]
        if distribution == 'Ubuntu':
            if '18.04' in version:
                version = '18.04'
            elif '20.04' in version:
                version = '20.04'
            elif '22.04' in version:
                version = '22.04'
            else:
                raise Exception(f"Ubuntu version {version} not recognized. Must be either '18.04', '20.04', or '22.04'.")
            dis_ver = f'{distribution}{version}'
            base_path = base_path+f'{dis_ver}/'
            tar_file = tar_file+f'{dis_ver}.tgz'
            download_link = base_path+tar_file
            print(base_path)
            print(tar_file)
            print(download_link)
            #test_link = 'https://heasarc.gsfc.nasa.gov/FTP/xmm/software/sparsebundle/components/DockerfileSAS'
            subprocess.run(['wget',download_link])
            subprocess.run(['tar','zxf',tar_file])
            subprocess.run(['./install.sh'],shell=True)
        else:
            raise Exception(f"Linux distribution {distribution} not supported.")

def generate_PPS_filename(obsid=None,instname=None,
                          expflag=None,expno=None,
                          product_type=None,datasubsetno=None,
                          sourceno=None,extension=None):
    """
    Function for generating a filename for downloading PPS files.

    PPS data product filenames take the 27.3 character form:

        POOOOOOOOOODDUEEETTTTTTSXXX.FFF

    P           The character P, to identify the files as a PPS product file

    OOOOOOOOOO  (obsid) Observation identifier 

    DD          (instname) Data source identifier (instrument name)

    U           (expflag) Exposure flag (1 character = S (sched),
                U (unsched), X (not applicable))

    EEE         (expno) Exposure number within the instrument 
                observation (3 digits)

    TTTTTT      (product_type) Product type (6 characters)

    S           (datasubsetno) 0 or data subset number/character 
                (1 character, differentiates energy bands,
                OSWs, filters, orders etc.)

    XXX         (sourceno) Source number or slew step number (3 characters, hexadecimal).
                It is set to 000 in source products from EPIC-pn Timing mode.

    FFF         (extension) File format (3 characters)

    If inputs are not given then a wildcard "*" character will 
    be inserted.

    """

    if not obsid: obsid = '*'
    if not instname: instname = '*'
    if not expflag: expflag = '*'
    if not expno: expno = '*'
    if not product_type: product_type = '*'
    if not datasubsetno: datasubsetno = '*'
    if not sourceno: sourceno = '*'
    if not extension: extension = '*'


    filename = f'P{obsid}{instname}{expflag}{expno}{product_type}{datasubsetno}{sourceno}.{extension}'

    re.sub(r'\*+', '*', filename)

    return filename


    
